"""STAR SHADOW
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This module contains functions for the use in and to perform
Markov Chain Monte Carlo (MCMC) with PyMC3

Code written by: Luc IJspeert
"""

import logging
import numpy as np
import scipy as sp
import scipy.stats

import star_shine.core.frequency_sets

try:
    # optional functionality
    import pymc3 as pm
    import theano.tensor as tt
    import arviz as az
    from fastprogress import fastprogress
except (ImportError, AttributeError) as e:
    if e == AttributeError:
        print('PyMC3 functionality unavailable, likely incompatible numpy version')
    pass

from star_shine.core import analysis as af


def sample_sinusoid(time, flux, const, slope, f_n, a_n, ph_n, c_err, sl_err, f_n_err, a_n_err, ph_n_err, noise_level,
                    i_chunks, logger=None):
    """NUTS sampling of a linear + sinusoid + eclipse model
    
    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves
    c_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the y-intercepts of a number of sine waves
    sl_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the slopes of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequencies of a number of sine waves
    a_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the amplitudes of a number of sine waves
    ph_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the phases of a number of sine waves
    noise_level: float
        The noise level (standard deviation of the residuals)
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    logger: logging.Logger, optional
        Instance of the logging library.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        inf_data: object
            Arviz inference data object
        par_means: list[float]
            Parameter mean values in the following order:
            const, slope, f_n, a_n, ph_n
        par_hdi: list[float]
            Parameter HDI error values, same order as par_means
    """
    # setup
    time_t = time.reshape(-1, 1)  # transposed time
    t_mean = tt.as_tensor_variable(np.mean(time))
    t_mean_s = tt.as_tensor_variable(np.array([np.mean(time[s[0]:s[1]]) for s in i_chunks]))
    lin_shape = (len(const),)
    sin_shape = (len(f_n),)
    # progress bar
    if logger is not None:
        fastprogress.printing = lambda: True
        mc_logger = logging.getLogger('pymc3')
        mc_logger.setLevel(logging.INFO)
    else:
        fastprogress.printing = lambda: False
        mc_logger = logging.getLogger('pymc3')
        mc_logger.setLevel(logging.ERROR)
    # make pymc3 model
    with pm.Model() as lc_model:
        # piece-wise linear curve parameter models
        const_pm = pm.Normal('const', mu=const, sigma=c_err, shape=lin_shape, testval=const)
        slope_pm = pm.Normal('slope', mu=slope, sigma=sl_err, shape=lin_shape, testval=slope)
        # piece-wise linear curve
        linear_curves = [const_pm[k] + slope_pm[k] * (time[s[0]:s[1]] - t_mean_s[k]) for k, s in enumerate(i_chunks)]
        model_linear = tt.concatenate(linear_curves)
        # sinusoid parameter models
        f_n_pm = pm.TruncatedNormal('f_n', mu=f_n, sigma=f_n_err, lower=0, shape=sin_shape, testval=f_n)
        a_n_pm = pm.TruncatedNormal('a_n', mu=a_n, sigma=a_n_err, lower=0, shape=sin_shape, testval=a_n)
        ph_n_pm = pm.VonMises('ph_n', mu=ph_n, kappa=1 / ph_n_err**2, shape=sin_shape, testval=ph_n)
        # sum of sinusoids
        model_sinusoid = pm.math.sum(a_n_pm * pm.math.sin((2 * np.pi * f_n_pm * (time_t - t_mean)) + ph_n_pm), axis=1)
        # full light curve model
        model = model_linear + model_sinusoid
        # observed distribution
        pm.Normal('obs', mu=model, sigma=noise_level, observed=flux)
    
    # do the sampling
    with lc_model:
        inf_data = pm.sample(draws=1000, tune=1000, init='adapt_diag', cores=1, progressbar=(logger is not None),
                             return_inferencedata=True)
    
    if logger is not None:
        az.summary(inf_data, round_to=2, circ_var_names=['ph_n'])
    # stacked parameter chains
    const_ch = inf_data.posterior.const.stack(dim=['chain', 'draw']).to_numpy()
    slope_ch = inf_data.posterior.slope.stack(dim=['chain', 'draw']).to_numpy()
    f_n_ch = inf_data.posterior.f_n.stack(dim=['chain', 'draw']).to_numpy()
    a_n_ch = inf_data.posterior.a_n.stack(dim=['chain', 'draw']).to_numpy()
    ph_n_ch = inf_data.posterior.ph_n.stack(dim=['chain', 'draw']).to_numpy()
    # parameter means
    const_m = np.mean(const_ch, axis=1).flatten()
    slope_m = np.mean(slope_ch, axis=1).flatten()
    f_n_m = np.mean(f_n_ch, axis=1).flatten()
    a_n_m = np.mean(a_n_ch, axis=1).flatten()
    ph_n_m = sp.stats.circmean(ph_n_ch, axis=1).flatten()
    par_means = [const_m, slope_m, f_n_m, a_n_m, ph_n_m]
    # parameter errors (from hdi) [hdi expects (chain, draw) as first two axes... annoying warnings...]
    const_e = az.hdi(np.moveaxis(const_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    slope_e = az.hdi(np.moveaxis(slope_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    f_n_e = az.hdi(np.moveaxis(f_n_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    a_n_e = az.hdi(np.moveaxis(a_n_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    ph_n_e = az.hdi(np.moveaxis(ph_n_ch[np.newaxis], 1, 2), hdi_prob=0.683, circular=True)
    # convert interval to error bars
    const_e = np.column_stack([const_m - const_e[:, 0], const_e[:, 1] - const_m])
    slope_e = np.column_stack([slope_m - slope_e[:, 0], slope_e[:, 1] - slope_m])
    f_n_e = np.column_stack([f_n_m - f_n_e[:, 0], f_n_e[:, 1] - f_n_m])
    a_n_e = np.column_stack([a_n_m - a_n_e[:, 0], a_n_e[:, 1] - a_n_m])
    ph_n_e = np.column_stack([ph_n_m - ph_n_e[:, 0], ph_n_e[:, 1] - ph_n_m])
    par_hdi = [const_e, slope_e, f_n_e, a_n_e, ph_n_e]
    return inf_data, par_means, par_hdi


def sample_sinusoid_h(time, flux, p_orb, const, slope, f_n, a_n, ph_n, p_err, c_err, sl_err, f_n_err, a_n_err,
                      ph_n_err, noise_level, i_chunks, logger=None):
    """NUTS sampling of a linear + sinusoid + eclipse model
    
    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    p_orb: float
        Orbital period of the eclipsing binary in days
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves
    p_err: float
        Uncertainty in the orbital period
    c_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the y-intercepts of a number of sine waves
    sl_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the slopes of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequencies of a number of sine waves
    a_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the amplitudes of a number of sine waves
    ph_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the phases of a number of sine waves
    noise_level: float
        The noise level (standard deviation of the residuals)
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    logger: logging.Logger, optional
        Instance of the logging library.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        inf_data: object
            Arviz inference data object
        par_means: list[float]
            Parameter mean values in the following order:
            p_orb, const, slope, f_n, a_n, ph_n
        par_hdi: list[float]
            Parameter HDI error values, same order as par_means
    """
    # setup
    time_t = time.reshape(-1, 1)  # transposed time
    t_mean = tt.as_tensor_variable(np.mean(time))
    t_mean_s = tt.as_tensor_variable(np.array([np.mean(time[s[0]:s[1]]) for s in i_chunks]))
    harmonics, harmonic_n = star_shine.core.frequency_sets.find_harmonics_from_pattern(f_n, 1/p_orb, f_tol=1e-9)
    non_harm = np.delete(np.arange(len(f_n)), harmonics)
    lin_shape = (len(const),)
    sin_shape = (len(f_n[non_harm]),)
    harm_shape = (len(f_n[harmonics]),)
    # progress bar
    if logger is not None:
        fastprogress.printing = lambda: True
        mc_logger = logging.getLogger('pymc3')
        mc_logger.setLevel(logging.INFO)
    else:
        fastprogress.printing = lambda: False
        mc_logger = logging.getLogger('pymc3')
        mc_logger.setLevel(logging.ERROR)
    # make pymc3 model
    with pm.Model() as lc_model:
        # piece-wise linear curve parameter models
        const_pm = pm.Normal('const', mu=const, sigma=c_err, shape=lin_shape, testval=const)
        slope_pm = pm.Normal('slope', mu=slope, sigma=sl_err, shape=lin_shape, testval=slope)
        # piece-wise linear curve
        linear_curves = [const_pm[k] + slope_pm[k] * (time[s[0]:s[1]] - t_mean_s[k]) for k, s in enumerate(i_chunks)]
        model_linear = tt.concatenate(linear_curves)
        # sinusoid parameter models
        f_n_pm = pm.TruncatedNormal('f_n', mu=f_n[non_harm], sigma=f_n_err[non_harm], lower=0, shape=sin_shape,
                                    testval=f_n[non_harm])
        a_n_pm = pm.TruncatedNormal('a_n', mu=a_n[non_harm], sigma=a_n_err[non_harm], lower=0, shape=sin_shape,
                                    testval=a_n[non_harm])
        ph_n_pm = pm.VonMises('ph_n', mu=ph_n[non_harm], kappa=1 / ph_n_err[non_harm]**2, shape=sin_shape,
                              testval=ph_n[non_harm])
        # sum of sinusoids
        model_sinusoid = pm.math.sum(a_n_pm * pm.math.sin((2 * np.pi * f_n_pm * (time_t - t_mean)) + ph_n_pm), axis=1)
        # harmonic parameter models
        p_orb_pm = pm.TruncatedNormal('p_orb', mu=p_orb, sigma=p_err, lower=0, testval=p_orb)
        f_h_pm = pm.Deterministic('f_h', harmonic_n / p_orb_pm)
        a_h_pm = pm.TruncatedNormal('a_h', mu=a_n[harmonics], sigma=a_n_err[harmonics], lower=0, shape=harm_shape,
                                    testval=a_n[harmonics])
        ph_h_pm = pm.VonMises('ph_h', mu=ph_n[harmonics], kappa=1 / ph_n_err[harmonics]**2, shape=harm_shape,
                              testval=ph_n[harmonics])
        # sum of harmonic sinusoids
        model_harmonic = pm.math.sum(a_h_pm * pm.math.sin((2 * np.pi * f_h_pm * (time_t - t_mean)) + ph_h_pm), axis=1)
        # full light curve model
        model = model_linear + model_sinusoid + model_harmonic
        # observed distribution
        pm.Normal('obs', mu=model, sigma=noise_level, observed=flux)
    
    # do the sampling
    with lc_model:
        inf_data = pm.sample(draws=1000, tune=1000, init='adapt_diag', cores=1, progressbar=(logger is not None),
                             return_inferencedata=True)
    
    if logger is not None:
        az.summary(inf_data, round_to=2, circ_var_names=['ph_n'])
    # stacked parameter chains
    p_orb_ch = inf_data.posterior.p_orb.stack(dim=['chain', 'draw']).to_numpy()
    const_ch = inf_data.posterior.const.stack(dim=['chain', 'draw']).to_numpy()
    slope_ch = inf_data.posterior.slope.stack(dim=['chain', 'draw']).to_numpy()
    f_n_ch = inf_data.posterior.f_n.stack(dim=['chain', 'draw']).to_numpy()
    a_n_ch = inf_data.posterior.a_n.stack(dim=['chain', 'draw']).to_numpy()
    ph_n_ch = inf_data.posterior.ph_n.stack(dim=['chain', 'draw']).to_numpy()
    f_h_ch = inf_data.posterior.f_h.stack(dim=['chain', 'draw']).to_numpy()
    a_h_ch = inf_data.posterior.a_h.stack(dim=['chain', 'draw']).to_numpy()
    ph_h_ch = inf_data.posterior.ph_h.stack(dim=['chain', 'draw']).to_numpy()
    # parameter means
    p_orb_m = np.mean(p_orb_ch)
    const_m = np.mean(const_ch, axis=1).flatten()
    slope_m = np.mean(slope_ch, axis=1).flatten()
    f_n_m = np.mean(f_n_ch, axis=1).flatten()
    a_n_m = np.mean(a_n_ch, axis=1).flatten()
    ph_n_m = sp.stats.circmean(ph_n_ch, axis=1).flatten()
    f_h_m = harmonic_n / p_orb_m  # taking the mean from the chain leads to slightly different results and is wrong
    a_h_m = np.mean(a_h_ch, axis=1).flatten()
    ph_h_m = sp.stats.circmean(ph_h_ch, axis=1).flatten()
    f_n_m = np.append(f_n_m, f_h_m)
    a_n_m = np.append(a_n_m, a_h_m)
    ph_n_m = np.append(ph_n_m, ph_h_m)
    par_means = [p_orb_m, const_m, slope_m, f_n_m, a_n_m, ph_n_m]
    # parameter errors (from hdi) [hdi expects (chain, draw) as first two axes... annoying warnings...]
    p_orb_e = az.hdi(p_orb_ch, hdi_prob=0.683)
    const_e = az.hdi(np.moveaxis(const_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    slope_e = az.hdi(np.moveaxis(slope_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    f_n_e = az.hdi(np.moveaxis(f_n_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    a_n_e = az.hdi(np.moveaxis(a_n_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    ph_n_e = az.hdi(np.moveaxis(ph_n_ch[np.newaxis], 1, 2), hdi_prob=0.683, circular=True)
    f_h_e = az.hdi(np.moveaxis(f_h_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    a_h_e = az.hdi(np.moveaxis(a_h_ch[np.newaxis], 1, 2), hdi_prob=0.683)
    ph_h_e = az.hdi(np.moveaxis(ph_h_ch[np.newaxis], 1, 2), hdi_prob=0.683, circular=True)
    f_n_e = np.append(f_n_e, f_h_e, axis=0)
    a_n_e = np.append(a_n_e, a_h_e, axis=0)
    ph_n_e = np.append(ph_n_e, ph_h_e, axis=0)
    # convert interval to error bars
    p_orb_e = np.array([p_orb_m - p_orb_e[0], p_orb_e[1] - p_orb_m])
    const_e = np.column_stack([const_m - const_e[:, 0], const_e[:, 1] - const_m])
    slope_e = np.column_stack([slope_m - slope_e[:, 0], slope_e[:, 1] - slope_m])
    f_n_e = np.column_stack([f_n_m - f_n_e[:, 0], f_n_e[:, 1] - f_n_m])
    a_n_e = np.column_stack([a_n_m - a_n_e[:, 0], a_n_e[:, 1] - a_n_m])
    ph_n_e = np.column_stack([ph_n_m - ph_n_e[:, 0], ph_n_e[:, 1] - ph_n_m])
    par_hdi = [p_orb_e, const_e, slope_e, f_n_e, a_n_e, ph_n_e]
    return inf_data, par_means, par_hdi



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the pipeline class that defines the analysis pipeline.

Code written by: Luc IJspeert
"""
import os
import time as systime
import numpy as np

from star_shine.api.data import Data
from star_shine.api.result import Result

from star_shine.core import analysis as ana, time_series as tms
from star_shine.core import frequency_sets as frs, periodogram as pdg, fitting as fit
from star_shine.core import goodness_of_fit as gof, mcmc as mcf, utility as ut
from star_shine.config.helpers import get_config, get_custom_logger


# load configuration
config = get_config()


class Pipeline:
    """A class to analyze light curve data.

    Handles the full analysis pipeline of Star Shine.

    Attributes
    ----------
    data: Data
        Instance of the Data class holding the light curve data.
    result: Result
        Instance of the Result class holding the parameters of the result.
    ts_model: TimeSeriesModel
        Instance of the TimeSeriesModel class holding the time series model.
    save_dir: str
        Root directory where the result files will be stored.
    save_subdir: str
        Sub-directory that is made to contain the save files.
    logger: Logger
        Instance of the logging library.
    """

    def __init__(self, data, save_dir='', logger=None):
        """Initialises the Pipeline object.

        Parameters
        ----------
        data: Data
            Instance of the Data class with the data to be analysed.
        save_dir: str, optional
            Root directory where result files will be stored. Added to the file name.
            If empty, it is loaded from config.

        Notes
        -----
        Creates a directory where all the analysis result files will be stored.
        """
        # check the input data
        if not isinstance(data, Data):
            raise ValueError("Input `data` should be a Data object.")

        # set the data and result objects
        self.data = data  # the data to be analysed
        self.result = Result()  # an empty result instance

        # set the time series object with the time series from data
        self.ts_model = tms.TimeSeriesModel.from_time_series(self.data.time_series)

        # the files will be stored here
        if save_dir == '':
            save_dir = config.save_dir
        self.save_dir = save_dir
        self.save_subdir = f"{self.data.target_id}_analysis"

        # for saving, make a folder if not there yet
        full_dir = os.path.join(self.save_dir, self.save_subdir)
        if not os.path.isdir(full_dir):
            os.mkdir(full_dir)  # create the subdir

        # initialise custom logger
        self.logger = logger or get_custom_logger(self.data.target_id, full_dir, config.verbose)

        return

    def extract_approx(self, f_approx):
        """Extract a sinusoid from the time series at an approximate frequency.

        Parameters
        ----------
        f_approx: float
            Approximate location of the frequency of maximum amplitude.
        """
        f, a, ph = ana.extract_approx(self.ts_model.time, self.ts_model.residual(), f_approx)

        # if identical frequency exists, assume this was unintentional
        if self.ts_model.sinusoid.n_sin > 0 and np.min(np.abs(f - self.ts_model.sinusoid.f_n)) < self.ts_model.pd_df:
            self.logger.warning("Existing identical frequency found.")
            return None

        # add the sinusoid
        self.ts_model.add_sinusoids(f, a, ph)
        self.ts_model.update_linear_model()

        # remove any frequencies that end up not making the statistical cut
        ana.reduce_sinusoids(self.ts_model, logger=self.logger)

        # update the TimeSeriesModel passing masks and uncertainties
        ana.select_sinusoids(self.ts_model, logger=self.logger)

        # print some useful info
        self.logger.info(f"N_f= {self.ts_model.sinusoid.n_sin}, BIC= {self.ts_model.bic():1.2f}, "
                         f"N_p= {self.ts_model.n_param} - Extracted f= {f:1.2f}", extra={'update': True})

        return None

    def remove_approx(self, f_approx):
        """Remove a sinusoid from the list at an approximate frequency.

        Parameters
        ----------
        f_approx: float
            Approximate location of the frequency to be removed.
        """
        # guard against empty frequency array
        if len(self.ts_model.sinusoid.f_n) == 0:
            return None

        index = np.argmin(np.abs(f_approx - self.ts_model.sinusoid.f_n))
        f_to_remove = self.ts_model.sinusoid.f_n[index]

        # if too far away, assume this was unintentional
        if abs(f_to_remove - f_approx) > 3 * self.ts_model.f_resolution:
            self.logger.warning("No close frequency to remove.")
            return None

        # remove the sinusoid
        self.ts_model.remove_sinusoids(index)

        # update the TimeSeriesModel passing masks and uncertainties
        ana.select_sinusoids(self.ts_model, logger=self.logger)

        # print some useful info
        self.logger.info(f"N_f= {self.ts_model.sinusoid.n_sin}, BIC= {self.ts_model.bic():1.2f}, "
                         f"N_p= {self.ts_model.n_param} - Removed f= {f_to_remove:1.2f}", extra={'update': True})

        return None

    def iterative_prewhitening(self, n_extract=0):
        """Iterative prewhitening of the input flux time series in the form of sine waves and a piece-wise linear curve.

        After extraction, a final check is done to see whether some frequencies are better removed or groups of
        frequencies are better replaced by one frequency.

        Continues from last results if frequency list is not empty.

        Parameters
        ----------
        n_extract: int, optional
            Maximum number of frequencies to extract. The stop criterion is still leading. Zero means as many as possible.

        Returns
        -------
        Result
            Instance of the Result class containing the analysis results
        """
        t_a = systime.time()
        self.logger.info(f"Iterative prewhitening starting.")

        # start by looking for more harmonics - only do this when n_extract set to 0
        if self.ts_model.sinusoid.n_base != 0 and n_extract == 0:
            ana.extract_harmonics(self.ts_model, config.bic_thr, logger=self.logger)

        # extract all frequencies with the iterative scheme
        ana.extract_sinusoids(self.ts_model, bic_thr=config.bic_thr, snr_thr=config.snr_thr,
                              stop_crit=config.stop_criterion, select=config.select_next,  n_extract=n_extract,
                              fit_each_step=config.optimise_step, replace_each_step=config.replace_step,
                              logger=self.logger)

        # remove any frequencies that end up not making the statistical cut
        ana.reduce_sinusoids(self.ts_model, logger=self.logger)

        # update the TimeSeriesModel passing masks and uncertainties
        ana.select_sinusoids(self.ts_model, logger=self.logger)

        # print some useful info
        t_b = systime.time()
        self.logger.info(f"N_f= {self.ts_model.sinusoid.n_sin}, BIC= {self.ts_model.bic():1.2f}, "
                         f"N_p= {self.ts_model.n_param} - Iterative prewhitening finished. "
                         f"Time taken= {t_b - t_a:1.1f}s.")

        return None

    def optimise_sinusoid(self):
        """Optimise the parameters of the sinusoid and linear model

        Returns
        -------
        Result
            Instance of the Result class containing the analysis results
        """
        t_a = systime.time()
        self.logger.info("Starting multi-sinusoid NL-LS optimisation.")

        # optimisation
        fit.fit_multi_sinusoid_grouped(self.ts_model, logger=self.logger)
            # # make model including everything to calculate noise level
            # resid = self.data.time_series.flux - self.ts_model.model_linear() - self.ts_model.model_sinusoid()
            # n_param = 2 * len(self.result.const) + 3 * len(self.result.f_n)
            # noise_level = ut.std_unb(resid, len(self.data.time_series.time) - n_param)
            #
            # # formal linear and sinusoid parameter errors
            # out_a = ut.formal_uncertainties(self.data.time_series.time, resid, self.data.time_series.flux_err,
            #                                 self.result.a_n, self.data.time_series.i_chunks)
            # c_err, sl_err, f_n_err, a_n_err, ph_n_err = out_a
            #
            # # do not include those frequencies that have too big uncertainty
            # include = (ph_n_err < 1 / np.sqrt(6))  # circular distribution for ph_n cannot handle these
            # f_n, a_n, ph_n = self.result.f_n[include], self.result.a_n[include], self.result.ph_n[include]
            # f_n_err, a_n_err, ph_n_err = f_n_err[include], a_n_err[include], ph_n_err[include]
            #
            # # Monte Carlo sampling of the model
            # out_b = mcf.sample_sinusoid(self.data.time_series.time, self.data.time_series.flux, self.result.const,
            #                             self.result.slope, f_n, a_n, ph_n, self.result.c_err, self.result.sl_err,
            #                             f_n_err, a_n_err, ph_n_err, noise_level, self.data.time_series.i_chunks,
            #                             logger=self.logger)
            # inf_data, par_mean, par_hdi = out_b
            # self.result.from_dict(c_hdi=par_hdi[0], sl_hdi=par_hdi[1], f_n_hdi=par_hdi[2], a_n_hdi=par_hdi[3],
            #                       ph_n_hdi=par_hdi[4])

        # remove any frequencies that end up not making the statistical cut
        ana.reduce_sinusoids(self.ts_model, logger=self.logger)

        # update the TimeSeriesModel passing masks and uncertainties
        ana.select_sinusoids(self.ts_model, logger=self.logger)

        # ut.save_inference_data(file_name, inf_data)  # currently not saved

        # print some useful info
        t_b = systime.time()
        self.logger.info(f"N_f= {self.ts_model.sinusoid.n_sin}, BIC= {self.ts_model.bic():1.2f}, "
                         f"N_p= {self.ts_model.n_param} - Optimisation complete. Time taken: {t_b - t_a:1.1f}s.")

        return None

    def add_base_harmonic(self, f_base):
        """Add a base harmonic frequency to the model."""
        # get amplitude and phase
        a, ph = pdg.scargle_ampl_phase_single(self.ts_model.time, self.ts_model.residual(), f_base)

        # add the sinusoid
        index = len(self.ts_model.sinusoid.f_n)
        self.ts_model.add_sinusoids(f_base, a, ph, h_base_new=index, h_mult_new=1)
        self.ts_model.update_linear_model()

        # update the TimeSeriesModel passing masks and uncertainties
        ana.select_sinusoids(self.ts_model, logger=self.logger)

        # print some useful info
        self.logger.info(f"N_f= {self.ts_model.sinusoid.n_sin}, BIC= {self.ts_model.bic():1.2f}, "
                         f"N_p= {self.ts_model.n_param} - Added f= {f_base:1.2f}", extra={'update': True})

        return None

    def couple_harmonics(self):
        """Find the orbital period and couple harmonic frequencies to the orbital period

        Returns
        -------
        Result
            Instance of the Result class containing the analysis results

        Notes
        -----
        Performs a global search for the base frequency, if unknown. If a frequency is given, it is locally
        refined for better performance. Needs an existing list of extracted sinusoids.

        Removes theoretical harmonic candidate frequencies within the frequency resolution, then extracts
        a single harmonic at the theoretical location.

        Removes any frequencies that end up not making the statistical cut.
        """
        t_a = systime.time()
        self.logger.info("Coupling the harmonic frequencies to the base frequency.")

        # check for non-coupled f_base
        h_base_unique, h_base_map = self.ts_model.sinusoid.get_h_base_map()
        for i_h in h_base_unique:
            # take the first f_base without a series
            if len(h_base_map[h_base_map == i_h]) > 1:
                f_base = self.ts_model.sinusoid.f_n[i_h]
                f_base = ana.refine_harmonic_base_frequency(f_base, self.ts_model)
                break
        else:
            # if no break encountered (no non-coupled f_base found), an f_base is searched for globally
            f_base = ana.find_harmonic_base_frequency(self.ts_model)

        if (t_over_p := self.ts_model.t_tot * f_base) > 1.1:
            # couple the harmonics to the period. likely removes more frequencies that need re-extracting
            ana.couple_harmonics(self.ts_model, f_base, logger=self.logger)

        # remove any frequencies that end up not making the statistical cut
        ana.reduce_sinusoids(self.ts_model, logger=self.logger)

        # update the TimeSeriesModel passing masks and uncertainties
        ana.select_sinusoids(self.ts_model, logger=self.logger)

        # print some useful info
        t_b = systime.time()
        f_base_err = self.ts_model.sinusoid.f_h_err[self.ts_model.sinusoid.get_f_index(f_base)]
        f_base_formatted = ut.float_to_str_scientific(f_base, f_base_err, error=True, brackets=True)
        self.logger.info(f"N_f= {self.ts_model.sinusoid.n_sin}, BIC= {self.ts_model.bic():1.2f}, "
                         f"N_p= {self.ts_model.n_param} - Harmonic frequencies coupled. f_base= {f_base_formatted}, "
                         f"1/f_base= {1/f_base:1.2f}. Time taken: {t_b - t_a:1.1f}s")

        # log if short time span or few harmonics
        if t_over_p < 1.1:
            self.logger.warning(f"Time-base over period is less than two: {t_over_p}.")
        elif self.ts_model.sinusoid.n_harm < 2:
            self.logger.warning(f"Not enough harmonics found: {self.ts_model.sinusoid.n_harm}.")

        return None

    def run(self):
        """Run the analysis pipeline on the given data.

        Runs a predefined sequence of analysis steps. Stops at the stage defined in the configuration file.
        Files are saved automatically at the end of each stage, taking into account the desired behaviour for
        overwriting.

        The followed recipe is:

        1) Extract all frequencies
            We start by extracting the frequency with the highest amplitude one by one,
            directly from the Lomb-Scargle periodogram until the BIC does not significantly
            improve anymore. This step involves a final cleanup of the frequencies.

        2) Multi-sinusoid NL-LS optimisation
            The sinusoid parameters are optimised with a non-linear least-squares method,
            using groups of frequencies to limit the number of free parameters.

        3) Measure the orbital period and couple the harmonic frequencies
            Global search done with combined phase dispersion, Lomb-Scargle and length/
            filling factor of the harmonic series in the list of frequencies.
            Then sets the frequencies of the harmonics to their new values, coupling them
            to the orbital period. This step involves a final cleanup of the frequencies.
            [Note: it is possible to provide a fixed period if it is already known.
            It will still be optimised]

        4) Attempt to extract additional frequencies
            The decreased number of free parameters (2 vs. 3), the BIC, which punishes for free
            parameters, may allow the extraction of more harmonics.
            It is also attempted to extract more frequencies like in step 1 again, now taking
            into account the presence of harmonics.
            This step involves a final cleanup of the frequencies.

        5) Multi-sinusoid NL-LS optimisation with coupled harmonics
            Optimise all sinusoid parameters with a non-linear least-squares method,
            using groups of frequencies to limit the number of free parameters
            and including the orbital period and the coupled harmonics.

        Returns
        -------
        Result
            Instance of the Result class containing the analysis results
        """
        # this list determines which analysis steps are taken and in what order
        step_names = ['iterative_prewhitening', 'optimise_sinusoid', 'couple_harmonics', 'iterative_prewhitening',
                      'optimise_sinusoid']

        # if we have predefined periods, repeat the harmonic steps
        harmonic_step_names = ['couple_harmonics', 'iterative_prewhitening', 'optimise_sinusoid']

        # run steps until config number
        if config.stop_at_stage != 0:
            step_names = step_names[:config.stop_at_stage]

        # tag the start of the analysis
        t_a = systime.time()
        self.logger.info("Start of analysis")

        # reset the time series object with the time series from data
        self.ts_model = tms.TimeSeriesModel.from_time_series(self.data.time_series)

        # run this sequence for each analysis step of the pipeline
        for step in range(len(step_names)):
            file_name = os.path.join(self.save_dir, self.save_subdir, f'{self.data.target_id}_result_{step + 1}.hdf5')

            # skip to next step if result exists and we were not planning to overwrite it
            if os.path.isfile(file_name) and not config.overwrite:
                self.load_result(file_name)  # load result for next step

                continue

            # do the analysis step
            analysis_step = getattr(self, step_names[step])
            analysis_step()

            # save the results if conditions are met
            if not os.path.isfile(file_name) or config.overwrite:
                self.save_result(file_name)

        # final message and timing
        t_b = systime.time()
        self.logger.info(f"End of analysis. Total time elapsed: {t_b - t_a:1.1f}s.")

        return None

    def load_result(self, file_name):
        """Load a result file and update the model.

        Parameters
        ----------
        file_name: str
            File name to load the results from
        """
        # load result from file
        self.result = Result.load(file_name, logger=self.logger)

        # set the TimeSeriesModel
        self.ts_model = self.result.to_time_series_model(self.ts_model)

        return None

    def save_result(self, file_name):
        """Update the result form model and save.

        Parameters
        ----------
        file_name: str
            File name to load the results from
        """
        # update the result instance and set the identifiers
        self.result.from_time_series_model(self.ts_model, target_id=self.data.target_id, data_id=self.data.data_id)

        # save result to file
        self.result.save(file_name)

        return None



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains logging and text output functions for the graphical user interface.

Code written by: Luc IJspeert
"""
import logging

from PySide6.QtCore import QObject, Signal

from star_shine.config import helpers as hlp


class QLogEmitter(QObject):
    """A signal emitter helper class."""
    # Define a signal that emits the log message
    log_signal = Signal(str, bool)  # message, update flag

    def __init__(self):
        super().__init__()


class QSignalHandler(logging.Handler):
    """A custom logging handler that writes log messages to a QTextEdit widget."""
    # Define a signal that emits the log message
    log_signal = Signal(str)

    def __init__(self, emitter):
        """Initialise custom logging handler.

        Parameters
        ----------
        emitter: QLogEmitter
            Object that handles the signal emitting.
        """
        super().__init__()
        self.emitter = emitter  # instance of LogEmitter

    def emit(self, record):
        """Emit a log message to the QTextEdit widget.

        Parameters
        ----------
        record: logging.LogRecord
            The log record to be formatted and emitted.
        """
        msg = self.format(record)
        update = getattr(record, 'update', False)  # get the extra kwarg out of record
        self.emitter.log_signal.emit(msg, update)


def get_custom_gui_logger(target_id, save_dir):
    """Create a custom logger for logging to file and to the gui.

    Parameters
    ----------
    target_id: str
        Identifier to use for the log file.
    save_dir: str
        folder to save the log file. If empty, no saving happens.

    Returns
    -------
    logging.Logger
        Customised logger object
    """
    # get the normal non-verbose logger
    logger = hlp.get_custom_logger(target_id, save_dir=save_dir, verbose=False)

    # add a different stream handler
    qlog_emitter = QLogEmitter()
    qsignal_handler = QSignalHandler(qlog_emitter)
    qsignal_handler.setLevel(logging.EXTRA)  # print everything with level 15 or above
    s_format = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    qsignal_handler.setFormatter(s_format)
    logger.addHandler(qsignal_handler)

    # expose the log_signal to be connected
    logger.log_signal = qsignal_handler.emitter.log_signal

    return logger



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python script is meant to be run before first use, it ensures that the Just-In-Time compiler has done its job.
If your own use case involves time series longer than a few thousand data points, this is strongly recommended.
If not, this is less important, but do keep in mind that the first run will be slower.

Code written by: Luc IJspeert
"""

import os
import importlib.resources
import star_shine as sts


# get the path to the test light curve
data_path_traversable = importlib.resources.files('star_shine.data')
target_id = 'sim_000_lc'
file = data_path_traversable.joinpath(target_id + '.dat').as_posix()
data_path = os.path.split(file)[0]
file_list = [file]

# initialise the data and pipeline
data = sts.Data.load_data(file_list, data_dir='', target_id=target_id, data_id='', logger=None)
pipeline = sts.Pipeline(data, save_dir=data_path, logger=None)
pipeline.run()



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This module contains io functions for reading and writing data and results.

Code written by: Luc IJspeert
"""

import os
import h5py
import numpy as np
import numba as nb
import pandas as pd
import astropy.io.fits as fits
try:
    import arviz as az  # optional functionality
except ImportError:
    az = None
    pass

from star_shine.config.helpers import get_config

# load configuration
config = get_config()


def load_data_hdf5(file_name, h5py_file_kwargs=None):
    """Load data from an hdf5 file and return it in a dictionary.

    Primarily for the api class Data.

    Parameters
    ----------
    file_name: str
        File name (including path) for loading the data.
    h5py_file_kwargs: dict, optional
        Keyword arguments for opening the h5py file.
        Example: {'locking': False}, for a drive that does not support locking.

    Returns
    -------
    Dict
        Dictionary of the data attributes and fields
    """
    # to avoid dict in function defaults
    if h5py_file_kwargs is None:
        h5py_file_kwargs = {}

    # add everything to a dict
    data_dict = {}

    # load the results from the file
    with h5py.File(file_name, 'r', **h5py_file_kwargs) as file:
        # file description
        data_dict['target_id'] = file.attrs['target_id']
        data_dict['data_id'] = file.attrs['data_id']
        data_dict['date_time'] = file.attrs['date_time']

        # original list of files
        data_dict['data_dir'] = file.attrs['data_dir']
        data_dict['file_list'] = np.copy(file['file_list'])

        # summary statistics
        data_dict['t_tot'] = file.attrs['t_tot']
        data_dict['t_mean'] = file.attrs['t_mean']
        data_dict['t_step'] = file.attrs['t_step']

        # the time series data
        data_dict['time'] = np.copy(file['time'])
        data_dict['flux'] = np.copy(file['flux'])
        data_dict['flux_err'] = np.copy(file['flux_err'])

        # additional information
        data_dict['i_chunks'] = np.copy(file['i_chunks'])
        data_dict['flux_counts_medians'] = np.copy(file['flux_counts_medians'])
        data_dict['t_mean_chunk'] = np.copy(file['t_mean_chunk'])

    return data_dict


def save_data_hdf5(file_name, data_dict):
    """Save data to an hdf5 file.

    Primarily for the api class Data.

    Parameters
    ----------
    file_name: str
        File name (including path) for saving the data.
    data_dict: dict
        Dictionary of the data attributes and fields

    Returns
    -------
    None
    """
    # file name must have hdf5 extension
    ext = os.path.splitext(os.path.basename(file_name))[1]
    if ext != '.hdf5':
        file_name = file_name.replace(ext, '.hdf5')

    # save to hdf5
    with h5py.File(file_name, 'w') as file:
        file.attrs['target_id'] = data_dict['target_id']
        file.attrs['data_id'] = data_dict['data_id']
        file.attrs['date_time'] = data_dict['date_time']

        # original list of files
        file.attrs['data_dir'] = data_dict['data_dir']  # original data directory
        file.create_dataset('file_list', data=data_dict['file_list'])
        file['file_list'].attrs['description'] = 'original list of files for the creation of this data file'

        # summary statistics
        file.attrs['t_tot'] = data_dict['t_tot']  # Total time base of observations
        file.attrs['t_mean'] = data_dict['t_mean']  # Time reference (zero) point of the full light curve
        file.attrs['t_step'] = data_dict['t_step']  # Median time step of observations
        file.create_dataset('t_mean_chunk', data=data_dict['t_mean_chunk'])
        file['t_mean_chunk'].attrs['unit'] = 'time unit of the data (often days)'
        file['t_mean_chunk'].attrs['description'] = 'time reference (zero) point of the each time chunk'

        # the time series data
        file.create_dataset('time', data=data_dict['time'])
        file['time'].attrs['unit'] = 'time unit of the data (often days)'
        file['time'].attrs['description'] = 'timestamps of the observations'
        file.create_dataset('flux', data=data_dict['flux'])
        file['flux'].attrs['unit'] = 'median normalised flux'
        file['flux'].attrs['description'] = 'normalised flux measurements of the observations'
        file.create_dataset('flux_err', data=data_dict['flux_err'])
        file['flux_err'].attrs['unit'] = 'median normalised flux'
        file['flux_err'].attrs['description'] = 'normalised error measurements in the flux'

        # additional information
        file.create_dataset('i_chunks', data=data_dict['i_chunks'])
        file['i_chunks'].attrs['description'] = 'pairs of indices indicating time chunks of the data'
        file.create_dataset('flux_counts_medians', data=data_dict['flux_counts_medians'])
        file['flux_counts_medians'].attrs['unit'] = 'raw flux counts'
        file['flux_counts_medians'].attrs['description'] = 'median flux level per time chunk'

    return None


def load_result_hdf5(file_name, h5py_file_kwargs=None):
    """Load results from an hdf5 file and return it in a dictionary.

    Primarily for the api class Result.

    Parameters
    ----------
    file_name: str
        File name (including path) for loading the results.
    h5py_file_kwargs: dict, optional
        Keyword arguments for opening the h5py file.
        Example: {'locking': False}, for a drive that does not support locking.

    Returns
    -------
    Dict
        Dictionary of the result attributes and fields
    """
    # to avoid dict in function defaults
    if h5py_file_kwargs is None:
        h5py_file_kwargs = {}

    # add everything to a dict
    result_dict = {}

    # load the results from the file
    with h5py.File(file_name, 'r', **h5py_file_kwargs) as file:
        # file description
        result_dict['target_id'] = file.attrs['target_id']
        result_dict['data_id'] = file.attrs['data_id']
        result_dict['date_time'] = file.attrs['date_time']

        # summary statistics
        result_dict['n_param'] = file.attrs['n_param']
        result_dict['bic'] = file.attrs['bic']
        result_dict['noise_level'] = file.attrs['noise_level']

        # linear model parameters
        # y-intercepts
        result_dict['const'] = np.copy(file['const'])
        result_dict['const_err'] = np.copy(file['const_err'])
        result_dict['const_hdi'] = np.copy(file['const_hdi'])
        # slopes
        result_dict['slope'] = np.copy(file['slope'])
        result_dict['slope_err'] = np.copy(file['slope_err'])
        result_dict['slope_hdi'] = np.copy(file['slope_hdi'])

        # sinusoid model parameters
        # frequencies
        result_dict['f_n'] = np.copy(file['f_n'])
        result_dict['f_n_err'] = np.copy(file['f_n_err'])
        result_dict['f_n_hdi'] = np.copy(file['f_n_hdi'])
        # amplitudes
        result_dict['a_n'] = np.copy(file['a_n'])
        result_dict['a_n_err'] = np.copy(file['a_n_err'])
        result_dict['a_n_hdi'] = np.copy(file['a_n_hdi'])
        # phases
        result_dict['ph_n'] = np.copy(file['ph_n'])
        result_dict['ph_n_err'] = np.copy(file['ph_n_err'])
        result_dict['ph_n_hdi'] = np.copy(file['ph_n_hdi'])
        # passing criteria
        result_dict['passing_sigma'] = np.copy(file['passing_sigma'])
        result_dict['passing_snr'] = np.copy(file['passing_snr'])

        # harmonic model
        result_dict['h_base'] = np.copy(file['h_base'])
        result_dict['h_mult'] = np.copy(file['h_mult'])
        result_dict['f_h_err'] = np.copy(file['f_h_err'])
        # passing criteria
        result_dict['passing_harmonic'] = np.copy(file['passing_harmonic'])

    return result_dict


def save_result_hdf5(file_name, result_dict):
    """Save results to an hdf5 file.

    Primarily for the api class Result.

    Parameters
    ----------
    file_name: str
        File name (including path) for saving the result.
    result_dict: dict
        Dictionary of the result attributes and fields

    Returns
    -------
    None
    """
    # file name must have hdf5 extension
    ext = os.path.splitext(os.path.basename(file_name))[1]
    if ext != '.hdf5':
        file_name = file_name.replace(ext, '.hdf5')

    # save to hdf5
    with h5py.File(file_name, 'w') as file:
        file.attrs['target_id'] = result_dict['target_id']
        file.attrs['data_id'] = result_dict['data_id']
        file.attrs['date_time'] = result_dict['date_time']
        file.attrs['n_param'] = result_dict['n_param']  # number of free parameters
        file.attrs['bic'] = result_dict['bic']  # Bayesian Information Criterion of the residuals
        file.attrs['noise_level'] = result_dict['noise_level']  # standard deviation of the residuals

        # the linear model
        # y-intercepts
        file.create_dataset('const', data=result_dict['const'])
        file['const'].attrs['unit'] = 'median normalised flux'
        file['const'].attrs['description'] = 'y-intercept per analysed sector'
        file.create_dataset('const_err', data=result_dict['const_err'])
        file['const_err'].attrs['unit'] = 'median normalised flux'
        file['const_err'].attrs['description'] = 'errors in the y-intercept per analysed sector'
        file.create_dataset('const_hdi', data=result_dict['const_hdi'])
        file['const_hdi'].attrs['unit'] = 'median normalised flux'
        file['const_hdi'].attrs['description'] = 'HDI for the y-intercept per analysed sector'

        # slopes
        file.create_dataset('slope', data=result_dict['slope'])
        file['slope'].attrs['unit'] = 'median normalised flux / d'
        file['slope'].attrs['description'] = 'slope per analysed sector'
        file.create_dataset('slope_err', data=result_dict['slope_err'])
        file['slope_err'].attrs['unit'] = 'median normalised flux / d'
        file['slope_err'].attrs['description'] = 'error in the slope per analysed sector'
        file.create_dataset('slope_hdi', data=result_dict['slope_hdi'])
        file['slope_hdi'].attrs['unit'] = 'median normalised flux / d'
        file['slope_hdi'].attrs['description'] = 'HDI for the slope per analysed sector'

        # the sinusoid model
        # frequencies
        file.create_dataset('f_n', data=result_dict['f_n'])
        file['f_n'].attrs['unit'] = '1 / d'
        file['f_n'].attrs['description'] = 'frequencies of a number of sinusoids'
        file.create_dataset('f_n_err', data=result_dict['f_n_err'])
        file['f_n_err'].attrs['unit'] = '1 / d'
        file['f_n_err'].attrs['description'] = 'errors in the frequencies of a number of sinusoids'
        file.create_dataset('f_n_hdi', data=result_dict['f_n_hdi'])
        file['f_n_hdi'].attrs['unit'] = '1 / d'
        file['f_n_hdi'].attrs['description'] = 'HDI for the frequencies of a number of sinusoids'

        # amplitudes
        file.create_dataset('a_n', data=result_dict['a_n'])
        file['a_n'].attrs['unit'] = 'median normalised flux'
        file['a_n'].attrs['description'] = 'amplitudes of a number of sinusoids'
        file.create_dataset('a_n_err', data=result_dict['a_n_err'])
        file['a_n_err'].attrs['unit'] = 'median normalised flux'
        file['a_n_err'].attrs['description'] = 'errors in the amplitudes of a number of sinusoids'
        file.create_dataset('a_n_hdi', data=result_dict['a_n_hdi'])
        file['a_n_hdi'].attrs['unit'] = 'median normalised flux'
        file['a_n_hdi'].attrs['description'] = 'HDI for the amplitudes of a number of sinusoids'

        # phases
        file.create_dataset('ph_n', data=result_dict['ph_n'])
        file['ph_n'].attrs['unit'] = 'radians'
        file['ph_n'].attrs['description'] = 'phases of a number of sinusoids, with reference point t_mean'
        file.create_dataset('ph_n_err', data=result_dict['ph_n_err'])
        file['ph_n_err'].attrs['unit'] = 'radians'
        file['ph_n_err'].attrs['description'] = 'errors in the phases of a number of sinusoids'
        file.create_dataset('ph_n_hdi', data=result_dict['ph_n_hdi'])
        file['ph_n_hdi'].attrs['unit'] = 'radians'
        file['ph_n_hdi'].attrs['description'] = 'HDI for the phases of a number of sinusoids'

        # sinusoid selection criteria
        file.create_dataset('passing_sigma', data=result_dict['passing_sigma'])
        file['passing_sigma'].attrs['description'] = 'sinusoids passing the sigma criterion'
        file.create_dataset('passing_snr', data=result_dict['passing_snr'])
        file['passing_snr'].attrs['description'] = 'sinusoids passing the signal to noise criterion'

        # harmonic model
        file.create_dataset('h_base', data=result_dict['h_base'])
        file['h_base'].attrs['description'] = 'index of the base harmonic frequency'
        file.create_dataset('h_mult', data=result_dict['h_mult'])
        file['h_mult'].attrs['description'] = 'multiplier of the base harmonic frequency'
        file.create_dataset('f_h_err', data=result_dict['f_h_err'])
        file['f_h_err'].attrs['description'] = 'errors in the harmonic frequencies'

        # harmonic selection criteria
        file.create_dataset('passing_harmonic', data=result_dict['passing_harmonic'])
        file['passing_harmonic'].attrs['description'] = 'harmonic sinusoids passing the sigma criterion'

    return None


def save_result_csv(file_name, result_dict):
    """Save results to several csv files.

    Primarily for the api class Result.

    Parameters
    ----------
    file_name: str
        File name (including path) for saving the result.
    result_dict: dict
        Dictionary of the result attributes and fields

    Returns
    -------
    None
    """
    # file extension
    ext = os.path.splitext(os.path.basename(file_name))[1]

    # linear model parameters
    data = np.column_stack((result_dict['const'], result_dict['const_err'],
                            result_dict['const_hdi'][:, 0], result_dict['const_hdi'][:, 1],
                            result_dict['slope'], result_dict['slope_err'],
                            result_dict['slope_hdi'][:, 0], result_dict['slope_hdi'][:, 1]))

    hdr = 'const, const_err, const_hdi_l, const_hdi_r, slope, slope_err, slope_hdi_l, slope_hdi_r'
    file_name_lin = file_name.replace(ext, '_linear.csv')
    np.savetxt(file_name_lin, data, delimiter=',', header=hdr)

    # sinusoid model parameters
    data = np.column_stack((result_dict['f_n'], result_dict['f_n_err'],
                            result_dict['f_n_hdi'][:, 0], result_dict['f_n_hdi'][:, 1],
                            result_dict['a_n'], result_dict['a_n_err'],
                            result_dict['a_n_hdi'][:, 0], result_dict['a_n_hdi'][:, 1],
                            result_dict['ph_n'], result_dict['ph_n_err'],
                            result_dict['ph_n_hdi'][:, 0], result_dict['ph_n_hdi'][:, 1],
                            result_dict['passing_sigma'], result_dict['passing_snr'],
                            result_dict['h_base'], result_dict['h_mult'],
                            result_dict['f_h_err'], result_dict['passing_harmonic']
                            ))

    hdr = ('f_n, f_n_err, f_n_hdi_l, f_n_hdi_r, a_n, a_n_err, a_n_hdi_l, a_n_hdi_r, '
           'ph_n, ph_n_err, ph_n_hdi_l, ph_n_hdi_r, passing_sigma, passing_snr, '
           'h_base, h_mult, f_h_err, passing_h')
    file_name_sin = file_name.replace(ext, '_sinusoid.csv')
    np.savetxt(file_name_sin, data, delimiter=',', header=hdr)

    # period and statistics
    names = ('n_param', 'bic', 'noise_level')
    stats = (result_dict['n_param'], result_dict['bic'], result_dict['noise_level'])

    desc = ['Number of free parameters', 'Bayesian Information Criterion of the residuals',
            'Standard deviation of the residuals']
    data = np.column_stack((names, stats, desc))
    hdr = f"{result_dict['target_id']}, {result_dict['data_id']}, Model statistics\nname, value, description"
    file_name_stats = file_name.replace(ext, '_stats.csv')
    np.savetxt(file_name_stats, data, delimiter=',', header=hdr, fmt='%s')

    return None


@nb.njit(cache=True)
def normalise_counts(flux_counts, flux_counts_err, i_chunks):
    """Median-normalises flux (counts or otherwise, should be positive) by
    dividing by the median.

    Parameters
    ----------
    flux_counts: numpy.ndarray[Any, dtype[float]]
        Flux measurement values in counts of the time series.
    flux_counts_err: numpy.ndarray[Any, dtype[float]]
        Errors in the flux measurements.
    i_chunks: numpy.ndarray[Any, dtype[float]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    tuple:
        flux_norm: numpy.ndarray[Any, dtype[float]]
            Normalised flux measurements
        flux_err_norm: numpy.ndarray[Any, dtype[float]]
            Normalised flux errors (zeros if flux_counts_err is None)
        medians: numpy.ndarray[Any, dtype[float]]
            Median flux counts per chunk

    Notes
    -----
    The result is positive and varies around one.
    The flux is processed per chunk.
    """
    medians = np.zeros(len(i_chunks))
    flux_norm = np.zeros(len(flux_counts))
    flux_err_norm = np.zeros(len(flux_counts))

    for i, ch in enumerate(i_chunks):
        medians[i] = np.median(flux_counts[ch[0]:ch[1]])
        flux_norm[ch[0]:ch[1]] = flux_counts[ch[0]:ch[1]] / medians[i]
        flux_err_norm[ch[0]:ch[1]] = flux_counts_err[ch[0]:ch[1]] / medians[i]

    return flux_norm, flux_err_norm, medians


def sort_chunks(chunk_sorter, i_chunks):
    """Sorts the time chunks based on chunk_sorter and updates the indices accordingly.

    Parameters
    ----------
    chunk_sorter: np.ndarray
        Sort indices of the time chunks based on their means
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    tuple
        A tuple containing the following elements:
        time_sorter: numpy.ndarray[Any, dtype[int]]
            Sort indices for the full array
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Updated pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    """
    # Update i_chunks to be in the sorted order
    sorted_i_chunks = i_chunks[chunk_sorter]

    # Create full index array corresponding to the sorted chunks
    time_sorter = np.concatenate([np.arange(ch[0], ch[1]) for ch in sorted_i_chunks])

    # update i_chunks to the full sorted time array
    sorted_chunk_len = sorted_i_chunks[:, 1] - sorted_i_chunks[:, 0]
    index_high = np.cumsum(sorted_chunk_len)
    index_low = np.append([0], index_high[:-1])
    i_chunks = np.vstack((index_low, index_high)).T

    return time_sorter, i_chunks


def load_csv_data(file_name):
    """Load in the data from a single csv file.

    Change column names in the config file.

    Parameters
    ----------
    file_name: str
        File name (including path) of the data.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        flux: numpy.ndarray[Any, dtype[float]]
            Measurement values of the time series
        flux_err: numpy.ndarray[Any, dtype[float]]
            Errors in the measurement values
    """
    # get the right columns with pandas
    col_names = [config.cn_time, config.cn_flux, config.cn_flux_err]
    df = pd.read_csv(file_name, usecols=col_names)

    # convert to numpy arrays
    time, flux, flux_err = df[col_names].values.T

    return time, flux, flux_err


def load_fits_data(file_name):
    """Load in the data from a single fits file.

    The SAP flux is Simple Aperture Photometry, the processed data can be PDC_SAP, KSP_SAP, or other
    depending on the data source. Change column names in the config file.

    Parameters
    ----------
    file_name: str
        File name (including path) of the data.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        flux: numpy.ndarray[Any, dtype[float]]
            Measurement values of the time series
        flux_err: numpy.ndarray[Any, dtype[float]]
            Errors in the measurement values
        qual_flags: numpy.ndarray[Any, dtype[int]]
            Integer values representing the quality of the
            data points. Zero means good quality.
        crowdsap: float
            Light contamination parameter (1-third_light)

    Notes
    -----
    Originally written for TESS data products, adapted to be flexible.
    """
    # grab the time series data
    with fits.open(file_name, mode='readonly') as hdul:
        # time stamps and flux measurements
        time = hdul[1].data[config.cf_time]
        flux = hdul[1].data[config.cf_flux]
        flux_err = hdul[1].data[config.cf_flux_err]

        # quality flags
        qual_flags = hdul[1].data[config.cf_quality]

        # get crowding numbers if found
        if 'CROWDSAP' in hdul[1].header.keys():
            crowdsap = hdul[1].header['CROWDSAP']
        else:
            crowdsap = -1

    return time, flux, flux_err, qual_flags, crowdsap


def load_light_curve(file_list, apply_flags=True):
    """Load in the data from a list of ((TESS specific) fits) files.

    Also stitches the light curves of each individual file together and normalises to the median.

    Parameters
    ----------
    file_list: list[str]
        A list of file names (including path) of the data.
    apply_flags: bool
        Whether to apply the quality flags to the time series data

    Returns
    -------
    tuple:
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        flux: numpy.ndarray[Any, dtype[float]]
            Measurement values of the time series
        flux_err: numpy.ndarray[Any, dtype[float]]
            Errors in the measurement values
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        medians: numpy.ndarray[Any, dtype[float]]
            Median flux counts per chunk
    """
    time = np.array([])
    flux = np.array([])
    flux_err = np.array([])
    qual_flags = np.array([])
    i_chunks = np.zeros((0, 2), dtype=int)

    # load the data in list order
    for file in file_list:
        # get the data from the file with one of the following methods
        if file.endswith('.fits') | file.endswith('.fit'):
            ti, fl, err, qf, cro = load_fits_data(file)
        elif file.endswith('.csv') & ('pd' in locals()):
            ti, fl, err = load_csv_data(file)
            qf = np.zeros(len(ti))
            cro = 1
        else:
            ti, fl, err = np.loadtxt(file, usecols=(0, 1, 2), unpack=True)
            qf = np.zeros(len(ti))
            cro = 1

        # keep track of the data belonging to each time chunk
        chunk_index = [[len(i_chunks), len(i_chunks) + len(ti)]]
        if config.halve_chunks:
            chunk_index = [[len(i_chunks), len(i_chunks) + len(ti) // 2],
                           [len(i_chunks) + len(ti) // 2, len(i_chunks) + len(ti)]]

        i_chunks = np.append(i_chunks, chunk_index, axis=0)

        # append all other data
        time = np.append(time, ti)
        flux = np.append(flux, fl)
        flux_err = np.append(flux_err, err)
        qual_flags = np.append(qual_flags, qf)

    # sort chunks by time
    t_start = time[i_chunks[:0]]
    if np.any(np.diff(t_start) < 0):
        chunk_sorter = np.argsort(t_start)  # sort on chunk start time
        time_sorter, i_chunks = sort_chunks(chunk_sorter, i_chunks)
        time = time[time_sorter]
        flux = flux[time_sorter]
        flux_err = flux_err[time_sorter]
        qual_flags = qual_flags[time_sorter]

    # apply quality flags
    if apply_flags:
        # convert quality flags to boolean mask
        quality = (qual_flags == 0)
        time = time[quality]
        flux = flux[quality]
        flux_err = flux_err[quality]

    # clean up (on time and flux)
    finite = np.isfinite(time) & np.isfinite(flux)
    time = time[finite].astype(np.float64)
    flux = flux[finite].astype(np.float64)
    flux_err = flux_err[finite].astype(np.float64)

    # median normalise
    flux, flux_err, medians = normalise_counts(flux, flux_err, i_chunks)

    return time, flux, flux_err, i_chunks, medians


def save_inference_data(file_name, inf_data):
    """Save the inference data object from Arviz/PyMC3

    Parameters
    ----------
    file_name: str
        File name (including path) for saving the results.
    inf_data: object
        Arviz inference data object

    Returns
    -------
    None
    """
    if inf_data is None:
        return None

    fn_ext = os.path.splitext(os.path.basename(file_name))[1]
    file_name_mc = file_name.replace(fn_ext, '_dists.nc4')
    inf_data.to_netcdf(file_name_mc)

    return None


def read_inference_data(file_name):
    """Read the inference data object from Arviz/PyMC3

    Parameters
    ----------
    file_name: str
        File name (including path) for saving the results.

    Returns
    -------
    object
        Arviz inference data object
    """
    fn_ext = os.path.splitext(os.path.basename(file_name))[1]
    file_name_mc = file_name.replace(fn_ext, '_dists.nc4')
    inf_data = az.from_netcdf(file_name_mc)

    return inf_data



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the result class for handling the analysis results.

Code written by: Luc IJspeert
"""
import os
import numpy as np

from star_shine.core import utility as ut
from star_shine.core import io
from star_shine.config.helpers import get_config


# load configuration
config = get_config()


class Result:
    """A class to handle analysis results.

    Attributes
    ----------
    target_id: str
        User defined identification number or name for the target under investigation.
    data_id: str
        User defined identification name for the dataset used.
    n_param: int
        Number of free parameters in the model.
    bic: float
        Bayesian Information Criterion of the residuals.
    noise_level: float
        The noise level (standard deviation of the residuals).
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve.
    const_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the constant for each time chunk.
    const_hdi: numpy.ndarray[Any, dtype[float]]
        HDI bounds for the constant for each time chunk.
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve.
    slope_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the slope for each time chunk.
    slope_hdi: numpy.ndarray[Any, dtype[float]]
        HDI bounds for the slope for each time chunk.
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sinusoids.
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequency for each sinusoid.
    f_n_hdi: numpy.ndarray[Any, dtype[float]]
        HDI bounds for the frequency for each sinusoid.
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sinusoids.
    a_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the amplitude for each sinusoid (these are identical).
    a_n_hdi: numpy.ndarray[Any, dtype[float]]
        HDI bounds for the amplitude for each sinusoid.
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sinusoids.
    ph_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the phase for each sinusoid.
    ph_n_hdi: numpy.ndarray[Any, dtype[float]]
        HDI bounds for the phase for each sinusoid.
    passing_sigma: numpy.ndarray[Any, dtype[bool]]
        Sinusoids that passed the sigma check.
    passing_snr: numpy.ndarray[Any, dtype[bool]]
        Sinusoids that passed the signal-to-noise check.
    h_base: numpy.ndarray[Any, dtype[int]]
        Indices of the base frequencies in _f_n for each of _f_n. -1 indicates a free sinusoid.
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each of _f_n. 0 indicates a free sinusoid.
    f_h_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequency for each harmonic sinusoid.
    passing_harmonic: numpy.ndarray[Any, dtype[bool]]
        Sinusoids that passed the harmonic check.
    """

    def __init__(self):
        """Initialises the Result object."""
        # descriptive
        self.target_id = ''
        self.data_id = ''

        # summary statistics
        self.n_param = -1
        self.bic = -1.
        self.noise_level = -1.

        # attribute lists per model
        self.linear_property_list = ['const', 'const_err', 'const_hdi', 'slope', 'slope_err', 'slope_hdi']
        self.sinusoid_property_list = ['f_n', 'f_n_err', 'f_n_hdi', 'a_n', 'a_n_err', 'a_n_hdi',
                                       'ph_n', 'ph_n_err', 'ph_n_hdi', 'passing_sigma', 'passing_snr',
                                       'h_base', 'h_mult', 'f_h_err', 'passing_harmonic']

        # linear model parameters
        # y-intercepts
        self.const = np.zeros((0,))
        self.const_err = np.zeros((0,))
        self.const_hdi = np.zeros((0, 2))
        # slopes
        self.slope = np.zeros((0,))
        self.slope_err = np.zeros((0,))
        self.slope_hdi = np.zeros((0, 2))

        # sinusoid model parameters
        # frequencies
        self.f_n = np.zeros((0,))
        self.f_n_err = np.zeros((0,))
        self.f_n_hdi = np.zeros((0, 2))
        # amplitudes
        self.a_n = np.zeros((0,))
        self.a_n_err = np.zeros((0,))
        self.a_n_hdi = np.zeros((0, 2))
        # phases
        self.ph_n = np.zeros((0,))
        self.ph_n_err = np.zeros((0,))
        self.ph_n_hdi = np.zeros((0, 2))
        # passing criteria
        self.passing_sigma = np.zeros((0,), dtype=bool)
        self.passing_snr = np.zeros((0,), dtype=bool)

        # harmonic model
        self.h_base = np.zeros((0,), dtype=int)
        self.h_mult = np.zeros((0,), dtype=int)
        self.f_h_err = np.zeros((0,))
        # passing criteria
        self.passing_harmonic = np.zeros((0,), dtype=bool)

        return

    def get_dict(self):
        """Make a dictionary of the attributes.

        Primarily for saving to file.

        Returns
        -------
        dict
            Dictionary of the result attributes and fields
        """
        # make a dictionary of the fields to be saved
        result_dict = dict()
        result_dict['target_id'] = self.target_id
        result_dict['data_id'] = self.data_id
        result_dict['date_time'] = ut.datetime_formatted()

        result_dict['n_param'] = self.n_param  # number of free parameters
        result_dict['bic'] = self.bic  # Bayesian Information Criterion of the residuals
        result_dict['noise_level'] = self.noise_level  # standard deviation of the residuals

        # the linear model
        for key in self.linear_property_list:
            result_dict[key] = getattr(self, key)

        # the sinusoid model
        for key in self.sinusoid_property_list:
            result_dict[key] = getattr(self, key)

        return result_dict

    def from_dict(self, **kwargs):
        """Fill in the attributes with results in a dictionary (or keyword arguments).

        Parameters
        ----------
        kwargs:
            Accepts any of the class attributes as keyword input and sets them accordingly.
        """
        # set any attribute that exists if it is in the kwargs
        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        return None

    def from_time_series_model(self, ts_model, target_id=None, data_id=None):
        """Fill in the Result attributes with results from a TimeSeriesModel.

        Parameters
        ----------
        ts_model: tms.TimeSeriesModel
            Instance of TimeSeriesModel containing the time series and model parameters.
        target_id: str
            User defined identification number or name for the target under investigation.
        data_id: str
            User defined identification name for the dataset used.
        """
        # descriptive
        self.target_id = target_id
        self.data_id = data_id

        # summary statistics
        self.n_param = ts_model.n_param
        self.bic = ts_model.bic()
        self.noise_level = ut.std_unb(ts_model.residual(), ts_model.n_time - ts_model.n_param)

        # linear model parameters
        for key in self.linear_property_list:
            if '_hdi' not in key:  # avoid hdi for now
                setattr(self, key, getattr(ts_model.linear, key))

        # sinusoid model parameters
        ts_model.remove_excluded()  # clean up before transfer
        for key in self.sinusoid_property_list:
            if '_hdi' not in key:  # avoid hdi for now
                setattr(self, key, getattr(ts_model.sinusoid, key))

        return None

    @classmethod
    def load(cls, file_name, h5py_file_kwargs=None, logger=None):
        """Load a result file in hdf5 format.

        Parameters
        ----------
        file_name: str
            File name to load the results from
        h5py_file_kwargs: dict, optional
            Keyword arguments for opening the h5py file.
            Example: {'locking': False}, for a drive that does not support locking.
        logger: logging.Logger, optional
            Instance of the logging library.

        Returns
        -------
        Result
            Instance of the Result class with the loaded results.
        """
        # guard for not existing file
        if not os.path.isfile(file_name):
            instance = cls()
            return instance

        # add everything to a dict
        result_dict = io.load_result_hdf5(file_name, h5py_file_kwargs=h5py_file_kwargs)

        # initiate the Results instance and fill in the results
        instance = cls()
        instance.from_dict(**result_dict)

        if logger is not None:
            logger.info(f"Loaded result file with target identifier: {result_dict['target_id']}, "
                        f"created on {result_dict['date_time']}. Data identifier: {result_dict['data_id']}.")

        return instance

    def to_time_series_model(self, ts_model):
        """Copy the parameters into an existing TimeSeriesModel instance.

        Parameters
        ----------
        ts_model: tms.TimeSeriesModel
            Instance of TimeSeriesModel containing the time series and model parameters.

        Returns
        -------
        tms.TimeSeriesModel
            Instance of TimeSeriesModel containing the time series and model parameters.
        """
        # linear model
        ts_model.set_linear_model(self.const, self.slope)

        # set the errors
        ts_model.linear._const_err = self.const_err
        ts_model.linear._slope_err = self.slope_err

        # update numbers
        ts_model.linear.update_n()

        # sinusoid model
        ts_model.set_sinusoids(self.f_n, self.a_n, self.ph_n, self.h_base, self.h_mult)

        # set the errors
        ts_model.sinusoid._f_n_err = self.f_n_err
        ts_model.sinusoid._a_n_err = self.a_n_err
        ts_model.sinusoid._ph_n_err = self.ph_n_err
        ts_model.sinusoid._f_h_err = self.f_h_err

        # set the criterion passing masks
        ts_model.sinusoid._passing_sigma = self.passing_sigma
        ts_model.sinusoid._passing_snr = self.passing_snr
        ts_model.sinusoid._passing_harmonic = self.passing_harmonic

        # update numbers
        ts_model.sinusoid.update_n()

        return ts_model

    def save(self, file_name):
        """Save the results to a file in hdf5 format.

        Parameters
        ----------
        file_name: str
            File name to save the results to
        """
        # get a dictionary of the fields to be saved
        result_dict = self.get_dict()

        # io module handles writing to file
        io.save_result_hdf5(file_name, result_dict)

        # save csv files if configured
        if config.save_ascii:
            self.save_as_csv(file_name)

        return None

    def save_as_csv(self, file_name):
        """Write multiple ascii csv files for human readability.

        Parameters
        ----------
        file_name: str
            File name to save the results to
        """
        # get a dictionary of the fields to be saved
        result_dict = self.get_dict()

        # io module handles writing to file
        io.save_result_csv(file_name, result_dict)

        return None



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This module contains some getter functions for configuration purposes.

Code written by: Luc IJspeert
"""

import os
import logging
import tomllib
import importlib.metadata
import importlib.resources

from star_shine.config import config as cnfg


def get_version():
    """Get the version of the code from the metadata, or the project file if that fails.

    Returns
    -------
    str
        The version of the code

    Raises
    ------
    FileNotFoundError
        In case the package metadata is not set and subsequently the pyproject.toml was not found
    """
    try:
        version = importlib.metadata.version('star_shine')

    except importlib.metadata.PackageNotFoundError:
        # Fallback to reading from pyproject.toml if not installed
        project_root = os.path.dirname(os.path.abspath(__file__))
        pyproject_path = os.path.join(project_root, '../../pyproject.toml')  # Adjust the path as needed

        try:
            with open(pyproject_path, 'rb') as f:
                pyproject_data = tomllib.load(f)

        except (FileNotFoundError, KeyError, tomllib.TOMLDecodeError) as e:
            raise FileNotFoundError("Could not find or parse version in pyproject.toml")

        version = pyproject_data['project']['version']

    return version


def get_config():
    """Use this function to get the configuration

    Returns
    -------
    Config
        The singleton instance of Config.
    """
    return cnfg.get_config()


def get_config_path():
    """Get the path to the configuration file

    Returns
    -------
    str
        Path to the config file
    """
    # Use importlib.resources to find the path
    config_path = cnfg.get_config_path()

    return config_path


def get_images_path():
    """Get the path to the data/images folder

    Returns
    -------
    str
        Path to the data/images folder
    """
    # Use importlib.resources to find the path
    images_path = str(importlib.resources.files('star_shine.data').joinpath('images'))

    return images_path


def get_mpl_stylesheet_path():
    """Get the path to the matplotlib stylesheet

    Returns
    -------
    str
        Path to the matplotlib stylesheet
    """
    # Use importlib.resources to find the path
    stylesheet_path = str(importlib.resources.files('star_shine.config').joinpath('mpl_stylesheet.dat'))

    return stylesheet_path


def add_logging_level(level_name, level_num, method_name=None, verbose=False):
    """Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    http://stackoverflow.com/q/2183233/2988730

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        if verbose:
            print(f'{level_name} already defined in logging module')
        return None
    if hasattr(logging, method_name):
        if verbose:
            print(f'{method_name} already defined in logging module')
        return None
    if hasattr(logging.getLoggerClass(), method_name):
        if verbose:
            print(f'{method_name} already defined in logger class')
        return None

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)
    def logToRoot(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, logForLevel)
    setattr(logging, method_name, logToRoot)

    return None


def get_custom_logger(target_id, save_dir, verbose):
    """Create a custom logger for logging to file and to stdout

    Parameters
    ----------
    target_id: str
        Identifier to use for the log file.
    save_dir: str
        Folder to save the log file. If empty, no saving happens.
    verbose: bool
        If set to True, information will be printed by the logger.

    Returns
    -------
    logging.Logger
        Customised logger object
    """
    # define a custom logging level
    add_logging_level('EXTRA', logging.INFO - 5, method_name=None)

    # customize the logger
    logger = logging.getLogger(target_id)  # make an instance of the logging library
    logger.setLevel(logging.EXTRA)  # set base activation level for logger

    # make formatters for the handlers
    s_format = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    f_format = logging.Formatter(fmt='%(asctime)s %(levelname)s %(module)s - %(funcName)s: %(message)s',
                                 datefmt='%Y-%m-%d %H:%M:%S')

    # remove existing handlers to avoid duplicate messages
    if logger.hasHandlers():
        logger.handlers.clear()

    # make stream handler
    if verbose:
        s_handler = logging.StreamHandler()  # for printing
        s_handler.setLevel(logging.EXTRA)  # print everything with level 15 or above
        s_handler.setFormatter(s_format)
        logger.addHandler(s_handler)

    # file handler
    if save_dir != '':
        log_name = os.path.join(save_dir, f'{target_id}.log')
        f_handler = logging.FileHandler(log_name, mode='a')  # for saving
        f_handler.setLevel(logging.INFO)  # save everything with level 20 or above
        f_handler.setFormatter(f_format)
        logger.addHandler(f_handler)

    return logger



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the graphical user interface.

Code written by: Luc IJspeert
"""
import os
import sys
import functools

import numpy as np
from PySide6.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QHBoxLayout, QGridLayout, QSplitter, QMenuBar
from PySide6.QtWidgets import QWidget, QLabel, QTextEdit, QLineEdit, QSpinBox, QFileDialog, QMessageBox, QPushButton
from PySide6.QtWidgets import QTableView, QHeaderView, QDialog, QFormLayout
from PySide6.QtGui import QAction, QFont, QScreen, QStandardItemModel, QStandardItem, QTextCursor, QIcon

from star_shine.core import utility as ut
from star_shine.api import Data, Pipeline, main
from star_shine.gui import gui_log, gui_plot, gui_analysis, gui_config
from star_shine.config import helpers as hlp


# load configuration
config = hlp.get_config()


class MainWindow(QMainWindow):
    """The main window of the Star Shine application.

    Contains a graphical user interface for loading data, performing analysis,
    displaying results, and visualizing plots.
    """

    def __init__(self):
        super().__init__()

        # Get screen dimensions
        screen = QApplication.primaryScreen()
        screen_size = screen.availableSize()
        h_size = int(screen_size.width() * config.h_size_frac)  # some fraction of the screen width
        v_size = int(screen_size.height() * config.v_size_frac)  # some fraction of the screen height

        # Set some window things
        self.setWindowTitle("Star Shine")
        self.setGeometry(100, 50, h_size, v_size)  # x, y, width, height

        # App icon
        icon_path = os.path.join(hlp.get_images_path(), 'Star_Shine_dark_simple_small_transparent.png')
        self.setWindowIcon(QIcon(icon_path))

        # save dir needs a default value
        # self.save_dir = config.save_dir
        if config.save_dir == '':
            config.save_dir = os.path.expanduser('~')
        self.save_subdir = ''

        # Set font size for the entire application
        font = QFont()
        font.setPointSize(11)
        QApplication.setFont(font)

        # Create the central widget and set the layout
        self._setup_central_widget()

        # Create the menu bar
        self._setup_menu_bar()

        # Add widgets to the layout
        self._add_widgets_to_layout()

        # custom gui-specific logger (will be reloaded and connected when data is loaded)
        self.logger = gui_log.get_custom_gui_logger('gui_logger', '')
        self.logger.log_signal.connect(self.append_text)

        # add the api classes for functionality
        self.pipeline = None
        self.pipeline_thread = None

    def _setup_central_widget(self):
        """Set up the central widget and its layout."""
        # Create a central widget
        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)

        # create a horizontal layout
        self.layout = QHBoxLayout()
        self.central_widget.setLayout(self.layout)

        # create a splitter
        self.splitter = QSplitter()
        self.central_widget.layout().addWidget(self.splitter)

        return None

    def _add_widgets_to_layout(self):
        """Add widgets to the main window layout."""
        # Left column: input and text output field
        left_column = self._create_left_column()
        self.splitter.addWidget(left_column)

        # Middle column: frequencies
        middle_column = self._create_middle_column()
        self.splitter.addWidget(middle_column)

        # right column: plot area
        right_column = self._create_right_column()
        self.splitter.addWidget(right_column)

        # Set initial sizes for each column
        h_size = self.width()
        self.splitter.setSizes([h_size*3//9, h_size*2//9, h_size*4//9])

        return None

    def _setup_menu_bar(self):
        """Set up the menu bar with file and info menus."""
        menu_bar = QMenuBar()
        self.setMenuBar(menu_bar)

        # Add "File" menu and set it up
        file_menu = menu_bar.addMenu("File")
        self._setup_file_menu(file_menu)

        # Add "View" menu and set it up
        view_menu = menu_bar.addMenu("View")
        self._setup_view_menu(view_menu)

        # Add "Info" menu
        info_menu = menu_bar.addMenu("Info")
        about_action = QAction("About", self)
        about_action.triggered.connect(self.show_about_dialog)
        info_menu.addAction(about_action)

        return None

    def _setup_file_menu(self, file_menu):
        """Set up the file menu."""

        # Add "Load Data" button to "File" menu
        load_data_action = QAction("Load Data", self)
        load_data_action.triggered.connect(self.load_data)
        file_menu.addAction(load_data_action)

        # Add "Save Data Object" button to "File" menu
        save_data_action = QAction("Save Data", self)
        save_data_action.triggered.connect(self.save_data)
        file_menu.addAction(save_data_action)

        # Add "Load Result Object" button to "File" menu
        load_result_action = QAction("Load Result", self)
        load_result_action.triggered.connect(self.load_result)
        file_menu.addAction(load_result_action)

        # Add "Save Result Object" button to "File" menu
        save_result_action = QAction("Save Result", self)
        save_result_action.triggered.connect(self.save_result)
        file_menu.addAction(save_result_action)

        # Add a horizontal separator
        file_menu.addSeparator()

        # Add "Set Save Location" button to "File" menu
        set_save_location_action = QAction("Set Save Location", self)
        set_save_location_action.triggered.connect(self.set_save_location)
        file_menu.addAction(set_save_location_action)

        # Add "Settings" action to open the settings dialog
        settings_action = QAction("Settings", self)
        settings_action.triggered.connect(self.show_settings_dialog)
        file_menu.addAction(settings_action)

        # Add "Settings" action to open the settings dialog
        export_settings_action = QAction("Export Settings", self)
        export_settings_action.triggered.connect(self.export_settings)
        file_menu.addAction(export_settings_action)

        # Add "Settings" action to open the settings dialog
        import_settings_action = QAction("Import Settings", self)
        import_settings_action.triggered.connect(self.import_settings)
        file_menu.addAction(import_settings_action)

        # Add a horizontal separator
        file_menu.addSeparator()

        # Add "Exit" button to "File" menu
        exit_action = QAction("Exit", self)
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)

        return None

    def _setup_view_menu(self, view_menu):
        """Set up the view menu."""
        # Add a horizontal separator
        # view_menu.addSeparator()

        # Add "Save Layout" button to "View" menu
        self.save_layout_action = QAction("Save Layout", self)
        # self.save_layout_action.triggered.connect()
        view_menu.addAction(self.save_layout_action)

        return None

    def _create_left_column(self):
        """Create and return the left column widget.

        Returns
        -------
        QWidget
            The left column widget containing input fields, buttons, and text output.
        """
        # create a vertical layout for in the left column of the main layout
        l_col_widget = QWidget()
        l_col_layout = QVBoxLayout(l_col_widget)

        # create a horizontal layout for info fields and buttons
        l_col_top_widget = QWidget()
        l_col_top_layout = QHBoxLayout(l_col_top_widget)

        # create the info grid
        info_widget = self._create_info_fields()
        l_col_top_layout.addWidget(info_widget, stretch=1)

        # create the button grid
        button_widget = self._create_buttons_grid()
        l_col_top_layout.addWidget(button_widget, stretch=1)

        # add widget to the left column layout
        l_col_layout.addWidget(l_col_top_widget)

        # Log area
        log_label = QLabel("Log:")
        l_col_layout.addWidget(log_label)

        self.text_field = QTextEdit()
        self.text_field.setReadOnly(True)  # Make the text edit read-only
        l_col_layout.addWidget(self.text_field)

        return l_col_widget

    def _create_middle_column(self):
        """Create and return the middle column widget.

        Returns
        -------
        QWidget
            The middle column widget containing a table view.
        """
        # create a vertical layout for in the middle column of the main layout
        m_col_widget = QWidget()
        m_col_layout = QVBoxLayout(m_col_widget)

        # add the data model formula above the table
        equation_str = "Model: flux = \u2211\u1D62 (a\u1D62 sin(2\u03C0f\u1D62t + \u03C6\u1D62)) + bt + c"
        formula_label = QLabel(equation_str)
        m_col_layout.addWidget(formula_label)

        # Create the table view and model
        self.table_view = QTableView()
        self.table_model = QStandardItemModel(0, 3)  # Start with 0 rows and 3 columns
        self.table_model.setHorizontalHeaderLabels(["Frequency", "Amplitude", "Phase"])
        self.table_view.setModel(self.table_model)

        # Connect the selection changed signal
        self.table_view.selectionModel().selectionChanged.connect(self.update_plots)

        # Set the horizontal header's stretch mode for each column
        h_header = self.table_view.horizontalHeader()
        h_header.setSectionResizeMode(QHeaderView.ResizeMode.Stretch)  # Stretch all columns proportionally

        # Set the vertical header's stretch mode for each row
        v_header = self.table_view.verticalHeader()
        v_header.setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)

        # Add the button and table view to the middle column layout
        m_col_layout.addWidget(self.table_view)

        return m_col_widget

    def _create_right_column(self):
        """Create and return the right column widget.

        Returns
        -------
        QWidget
            The right column widget containing plot areas.
        """
        # create a vertical layout for in the right column of the main layout
        r_col_widget = QWidget()
        r_col_layout = QVBoxLayout(r_col_widget)

        # upper plot area for the data
        self.upper_plot_area = gui_plot.PlotWidget(title='Data', xlabel='time', ylabel='flux')
        r_col_layout.addWidget(self.upper_plot_area)

        # connect the residual event
        self.upper_plot_area.residual_signal.connect(functools.partial(self.update_plots, new_plot=True))

        # lower plot area for the periodogram
        self.lower_plot_area = gui_plot.PlotWidget(title='Periodogram', xlabel='frequency', ylabel='amplitude')
        r_col_layout.addWidget(self.lower_plot_area)

        # connect the click and residual event
        self.lower_plot_area.click_signal.connect(self.click_periodogram)
        self.lower_plot_area.residual_signal.connect(functools.partial(self.update_plots, new_plot=True))

        return r_col_widget

    def _create_buttons_grid(self):
        """Create and return the left column button grid widget.

        Returns
        -------
        QWidget
            The left column button grid widget.
        """
        # create a grid area for labels and read-only text fields
        button_widget = QWidget()
        button_layout = QGridLayout(button_widget)

        # Create a spin box for integer input
        self.spin_box = QSpinBox(self)
        self.spin_box.setRange(0, 99999)
        self.spin_box.setValue(0)

        # Button for starting iterative prewhitening
        extract_button = QPushButton("Extract")
        extract_button.clicked.connect(lambda: self.perform_analysis('iterative_prewhitening',
                                                                     n_extract=self.spin_box.value()))
        button_layout.addWidget(extract_button, 0, 0)
        button_layout.addWidget(self.spin_box, 0, 1)  # add the number field after the button

        # Button for starting sinusoid fit
        optimise_button = QPushButton("Optimise")
        optimise_button.clicked.connect(functools.partial(self.perform_analysis, 'optimise_sinusoid'))
        button_layout.addWidget(optimise_button, 1, 0)

        # Button for inputting a base harmonic
        add_harmonic_button = QPushButton("Add Base Harmonic")
        add_harmonic_button.clicked.connect(self.add_base_harmonic)
        button_layout.addWidget(add_harmonic_button, 2, 0)

        # Button for coupling harmonics
        couple_harmonic_button = QPushButton("Couple Harmonics")
        couple_harmonic_button.clicked.connect(functools.partial(self.perform_analysis, 'couple_harmonics'))
        button_layout.addWidget(couple_harmonic_button, 2, 1)

        # Button for adding a sinusoid manually
        add_button = QPushButton("Add Sinusoid")
        add_button.clicked.connect(self.add_sinusoid)
        button_layout.addWidget(add_button, 3, 0)

        # Button for deleting a selected sinusoid
        delete_button = QPushButton("Delete Sinusoid")
        delete_button.clicked.connect(self.delete_sinusoid)
        button_layout.addWidget(delete_button, 3, 1)

        # Button for saving the results
        save_result_button = QPushButton("Save Result")
        save_result_button.clicked.connect(self.save_result)
        button_layout.addWidget(save_result_button, 4, 0)

        # Button for starting analysis
        analyze_button = QPushButton("Run Full Analyis")
        analyze_button.clicked.connect(functools.partial(self.perform_analysis, 'run'))
        button_layout.addWidget(analyze_button, 5, 0)

        # Button for starting analysis
        interrupt_button = QPushButton("Interrupt [WIP]")
        interrupt_button.clicked.connect(self.stop_analysis)
        button_layout.addWidget(interrupt_button, 5, 1)

        return button_widget

    def _create_info_fields(self):
        """Create and return the left column information grid widget.

        Returns
        -------
        QWidget
            The left column information grid widget.
        """
        # create a grid area for labels and read-only text fields
        info_widget = QWidget()
        info_layout = QGridLayout(info_widget)

        # current directory
        current_dir_label = QLabel("Current directory:")
        self.current_dir_field = QLineEdit()
        self.current_dir_field.setReadOnly(True)
        self.current_dir_field.setText(config.save_dir)
        info_layout.addWidget(current_dir_label, 0, 0)
        info_layout.addWidget(self.current_dir_field, 0, 1)

        # target id
        target_id_label = QLabel("Target ID:")
        self.target_id_field = QLineEdit()
        self.target_id_field.setReadOnly(True)
        info_layout.addWidget(target_id_label, 1, 0)
        info_layout.addWidget(self.target_id_field, 1, 1)

        # data id
        data_id_label = QLabel("Data ID:")
        self.data_id_field = QLineEdit()
        self.data_id_field.setReadOnly(True)
        info_layout.addWidget(data_id_label, 2, 0)
        info_layout.addWidget(self.data_id_field, 2, 1)

        # Number of data points in the time series
        n_time_label = QLabel("N data points:")
        self.n_time_field = QLineEdit()
        self.n_time_field.setReadOnly(True)
        info_layout.addWidget(n_time_label, 3, 0)
        info_layout.addWidget(self.n_time_field, 3, 1)

        # Number of time chunks in the time series
        n_chunks_label = QLabel("N time chunks:")
        self.n_chunks_field = QLineEdit()
        self.n_chunks_field.setReadOnly(True)
        info_layout.addWidget(n_chunks_label, 4, 0)
        info_layout.addWidget(self.n_chunks_field, 4, 1)

        # Total time base of observations
        t_tot_label = QLabel("Total time base:")
        self.t_tot_field = QLineEdit()
        self.t_tot_field.setReadOnly(True)
        info_layout.addWidget(t_tot_label, 5, 0)
        info_layout.addWidget(self.t_tot_field, 5, 1)

        # Time reference (zero) point of the full light curve
        t_mean_label = QLabel("Time reference point:")
        self.t_mean_field = QLineEdit()
        self.t_mean_field.setReadOnly(True)
        info_layout.addWidget(t_mean_label, 6, 0)
        info_layout.addWidget(self.t_mean_field, 6, 1)

        # Median time step of observations
        t_step_label = QLabel("Median time step:")
        self.t_step_field = QLineEdit()
        self.t_step_field.setReadOnly(True)
        info_layout.addWidget(t_step_label, 7, 0)
        info_layout.addWidget(self.t_step_field, 7, 1)

        # Frequency resolution of the time series
        f_resolution_label = QLabel("Frequency resolution:")
        self.f_resolution_field = QLineEdit()
        self.f_resolution_field.setReadOnly(True)
        info_layout.addWidget(f_resolution_label, 8, 0)
        info_layout.addWidget(self.f_resolution_field, 8, 1)

        return info_widget

    def update_info_fields(self):
        """Update the information field widget."""
        # current directory
        self.current_dir_field.setText(config.save_dir)

        # if there is a pipeline we can fill in the data attributes
        if self.pipeline is not None:
            # update any properties in the TimeSeries that might have changed
            self.pipeline.ts_model.update_properties()

            self.target_id_field.setText(self.pipeline.data.target_id)
            self.data_id_field.setText(self.pipeline.data.data_id)
            self.n_time_field.setText(str(self.pipeline.ts_model.n_time))
            self.n_chunks_field.setText(str(self.pipeline.ts_model.n_chunks))
            self.t_tot_field.setText(str(self.pipeline.ts_model.t_tot))
            self.t_mean_field.setText(str(self.pipeline.ts_model.t_mean))
            self.t_step_field.setText(str(self.pipeline.ts_model.t_step))
            self.f_resolution_field.setText(str(self.pipeline.ts_model.f_resolution))

        return None

    def append_text(self, text):
        """Append a line of text at the end of the plain text output box.

        Parameters
        ----------
        text: str
            The text to append.
        """
        cursor = self.text_field.textCursor()
        cursor.movePosition(QTextCursor.End)  # Move cursor to the end of the text
        cursor.insertText(text + '\n')  # insert the text
        self.text_field.setTextCursor(cursor)
        self.text_field.ensureCursorVisible()

        return None

    def update_table(self, display_err=True):
        """Fill the table with the given data."""
        # get the result parameters
        col1 = self.pipeline.ts_model.sinusoid.f_n
        col2 = self.pipeline.ts_model.sinusoid.a_n
        col3 = self.pipeline.ts_model.sinusoid.ph_n

        # make sure uncertainties are updated
        self.pipeline.ts_model.update_sinusoid_uncertainties()
        col1_err = self.pipeline.ts_model.sinusoid.f_n_err
        col2_err = self.pipeline.ts_model.sinusoid.a_n_err
        col3_err = self.pipeline.ts_model.sinusoid.ph_n_err

        # display sinusoid parameters in the table
        self.table_model.setRowCount(len(col1))
        for row, row_items in enumerate(zip(col1, col2, col3, col1_err, col2_err, col3_err)):
            # convert to strings
            c1 = ut.float_to_str_scientific(row_items[0], row_items[3], error=display_err, brackets=False)
            c2 = ut.float_to_str_scientific(row_items[1], row_items[4], error=display_err, brackets=False)
            c3 = ut.float_to_str_scientific(row_items[2], row_items[5], error=display_err, brackets=False)

            # convert to table items
            c1_item = QStandardItem(c1)
            c2_item = QStandardItem(c2)
            c3_item = QStandardItem(c3)

            # insert into table
            self.table_model.setItem(row, 0, c1_item)
            self.table_model.setItem(row, 1, c2_item)
            self.table_model.setItem(row, 2, c3_item)

        return None

    def update_plots(self, new_plot=False):
        """Update the plotting area with the current data."""
        # collect plot data in a dict
        upper_plot_data = {}
        lower_plot_data = {}

        # collect attributes
        n_param = self.pipeline.ts_model.sinusoid.n_param
        time = self.pipeline.ts_model.time
        flux = self.pipeline.ts_model.flux

        # Get the row numbers of the selected indexes (if any)
        selected_rows = np.unique([index.row() for index in self.table_view.selectedIndexes()])

        # upper plot area - time series
        upper_plot_data['scatter_xs'] = [time]
        upper_plot_data['scatter_ys'] = [flux]
        # lower plot area - periodogram
        lower_plot_data['plot_xs'] = [self.pipeline.ts_model.pd_freqs]
        lower_plot_data['plot_ys'] = [self.pipeline.ts_model.pd_ampls]

        # include result attributes if present
        if n_param > 0:
            # calculate model and residual
            model = self.pipeline.ts_model.calc_model()
            residual = flux - model

            # calculate periodogram
            freqs, ampls = self.pipeline.ts_model.calc_periodogram()

            # upper plot area - time series
            if not self.upper_plot_area.show_residual:
                upper_plot_data['plot_xs'] = [time]
                upper_plot_data['plot_ys'] = [model]
                upper_plot_data['plot_colors'] = ['grey']

                # if highlighted rows
                if len(selected_rows) > 0:
                    highlighted_model = self.pipeline.ts_model.calc_model(indices=selected_rows)
                    upper_plot_data['plot_xs'].append(time)
                    upper_plot_data['plot_ys'].append(highlighted_model)
                    upper_plot_data['plot_colors'].append('tab:red')
            else: # only show residual if toggle checked
                upper_plot_data['scatter_xs'] = [time]
                upper_plot_data['scatter_ys'] = [residual]

            # lower plot area - periodogram
            if not self.lower_plot_area.show_residual:
                lower_plot_data['plot_xs'].append(freqs)
                lower_plot_data['plot_ys'].append(ampls)
                lower_plot_data['vlines_xs'] = [self.pipeline.ts_model.sinusoid.f_n]
                lower_plot_data['vlines_ys'] = [self.pipeline.ts_model.sinusoid.a_n]
                lower_plot_data['vlines_colors'] = ['grey']

                # if highlighted rows
                if len(selected_rows) > 0:
                    lower_plot_data['vlines_xs'].append(self.pipeline.ts_model.sinusoid.f_n[selected_rows])
                    lower_plot_data['vlines_ys'].append(self.pipeline.ts_model.sinusoid.a_n[selected_rows])
                    lower_plot_data['vlines_colors'].append('tab:red')
            else: # only show residual if toggle checked
                lower_plot_data['plot_xs'] = [freqs]
                lower_plot_data['plot_ys'] = [ampls]

        # set the plot data
        self.upper_plot_area.set_plot_data(**upper_plot_data)
        self.lower_plot_area.set_plot_data(**lower_plot_data)

        # start with a fresh plot
        if new_plot:
            self.upper_plot_area.new_plot()
            self.lower_plot_area.new_plot()
        else:
            self.upper_plot_area.update_plot()
            self.lower_plot_area.update_plot()

        return None

    def on_result_update(self, msg=None, update=False, new_plot=False, display_err=True):
        """Update the GUI with the results."""
        # show the message in the log area
        if msg is not None:
            self.append_text(msg)

        if update:
            # display sinusoid parameters in the table
            self.update_table(display_err=display_err)

            # Update the plot area with the results (or clear it first)
            self.update_plots(new_plot=new_plot)

        return None

    def new_dataset(self, data):
        """Set up pipeline and logger for the new data that was loaded"""
        # for saving, make a folder if not there yet
        self.save_subdir = f"{data.target_id}_analysis"

        full_dir = os.path.join(config.save_dir, self.save_subdir)
        if not os.path.isdir(full_dir):
            os.mkdir(full_dir)  # create the subdir

        # custom gui-specific logger
        self.logger = gui_log.get_custom_gui_logger(data.target_id, full_dir)
        self.logger.log_signal.connect(self.on_result_update)

        # Make ready the pipeline class
        self.pipeline = Pipeline(data=data, save_dir=config.save_dir, logger=self.logger)

        # set up a pipeline thread
        self.pipeline_thread = gui_analysis.PipelineThread(self.pipeline)

        # update the info fields
        self.update_info_fields()

        # display sinusoid parameters in the table and clear and update the plots
        self.on_result_update(msg=None, update=True, new_plot=True, display_err=True)

        return None

    def set_save_location(self):
        """Open a dialog to select the save location."""
        # Open a directory selection dialog
        new_dir = QFileDialog.getExistingDirectory(self, caption="Select Save Location", dir=config.save_dir)

        if new_dir:
            config.save_dir = new_dir
            self.logger.info(f"Save location set to: {config.save_dir}")

        # update the current directory info field
        self.update_info_fields()

        return None

    def load_data(self):
        """Read data from a file or multiple files using a dialog window."""
        # get the path(s) from a standard file selection screen
        file_paths, _ = QFileDialog.getOpenFileNames(self, caption="Read Data", dir=config.save_dir,
                                                     filter="All Files (*)")

        # do nothing in case no file(s) selected
        if not file_paths:
            return None

        # load data into instance
        if len(file_paths) == 1 and file_paths[0].endswith('.hdf5'):
            # a single hdf5 file is loaded as a star shine data object
            data = Data.load(file_name=file_paths[0], data_dir='', logger=self.logger)
        else:
            # any other files are loaded as external data
            data = Data.load_data(file_list=file_paths, data_dir='', target_id='', data_id='', logger=self.logger)

        # set the save dir to the one where we opened the data
        config.save_dir = os.path.dirname(file_paths[0])

        # set up some things
        self.new_dataset(data)

        return None

    def save_data(self):
        """Save data to a file using a dialog window."""
        # check whether data is present
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: load data first.")
            return None

        suggested_path = os.path.join(config.save_dir, self.pipeline.data.target_id + '_data.hdf5')
        file_path, _ = QFileDialog.getSaveFileName(self, caption="Save Data", dir=suggested_path,
                                                   filter="HDF5 Files (*.hdf5);;All Files (*)")

        # do nothing in case no file selected
        if not file_path:
            return None

        self.pipeline.data.save(file_path)

        return None

    def load_result(self):
        """Load result from a file using a dialog window."""
        # check whether a pipeline is present
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: load data first.")
            return None

        # get the path(s) from a standard file selection screen
        file_path, _ = QFileDialog.getOpenFileName(self, caption="Load Result", dir=config.save_dir,
                                                    filter="HDF5 Files (*.hdf5);;All Files (*)")

        # do nothing in case no file selected
        if not file_path:
            return None

        # load result into instance
        try:
            self.pipeline.load_result(file_path)
        except KeyError as e:
            self.logger.error(f"Incompatible file: {e}")

        # display sinusoid parameters in the table and clear and update the plots
        self.on_result_update(msg=None, update=True, new_plot=False, display_err=True)

        return None

    def save_result(self):
        """Save result to a file using a dialog window."""
        # check whether a result is present
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: load data first.")
            return None

        suggested_path = os.path.join(config.save_dir, self.pipeline.data.target_id + '_result.hdf5')
        file_path, _ = QFileDialog.getSaveFileName(self, caption="Save Data", dir=suggested_path,
                                                   filter="HDF5 Files (*.hdf5);;All Files (*)")

        # do nothing in case no file selected
        if not file_path:
            return None

        self.pipeline.save_result(file_path)

        return None

    def add_base_harmonic(self):
        """Let the user add a base harmonic frequency and add the harmonic series."""
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: Load data first.")
            return None

        # get some parameters
        min_val = self.pipeline.ts_model.pd_f0
        max_val = self.pipeline.ts_model.pd_fn
        text = f"Enter a number between {min_val:1.2f} and {max_val:1.2f}."

        # open the dialog
        dialog = InputDialog("Add base harmonic frequency", text)

        if dialog.exec():
            value = dialog.get_values()[0]

            try:
                value = float(value)
            except ValueError:
                QMessageBox.warning(self, "Warning", "Invalid input: not a float.")

                return None

            if value < min_val or value > max_val:
                QMessageBox.warning(self, "Warning", "Value out of range.")

                return None

            # if we made it here, add the harmonics
            self.pipeline_thread.start_function('add_base_harmonic', value)

        return None

    def perform_analysis(self, func_name, *args, **kwargs):
        """Perform analysis on the loaded data and display results."""
        # check whether data is loaded
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: provide data files.")
            return None

        # start a new thread for the analysis
        self.pipeline_thread.start_function(func_name, *args, **kwargs)

        return None

    def stop_analysis(self):
        """Stop the analysis, if it is running."""
        if self.pipeline_thread is not None:
            self.pipeline_thread.stop()

        return None

    def click_periodogram(self, x, y, button):
        """Handle click events on the periodogram plot."""
        # Guard against empty data
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.info(f"Plot clicked at coordinates: ({x}, {y})")
            return None

        # Left click
        if button == 1:
            self.pipeline_thread.start_function('extract_approx', x)

        # Right click
        if button == 3:
            self.pipeline_thread.start_function('remove_approx', x)

        return None

    def add_sinusoid(self):
        """Manually add a sinusoid to the model."""
        if self.pipeline is None or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: load data first.")
            return None

        # make the text fields
        text_1 = "Frequency:"
        text_2 = "Amplitude:"
        text_3 = "Phase:"

        # open the dialog
        dialog = InputDialog("Add sinusoid", text_1, text_2, text_3)

        if dialog.exec():
            values = dialog.get_values()

            try:
                value_1 = float(values[0])
                value_2 = float(values[1])
                value_3 = float(values[2])
            except ValueError:
                QMessageBox.warning(self, "Warning", "Invalid input: not a float.")

                return None

            # add the sinusoid
            self.pipeline_thread.start_function('add_sinusoid', value_1, value_2, value_3)

        return None

    def delete_sinusoid(self):
        """Delete the sinusoid(s) selected in the list."""
        if self.pipeline is None  or len(self.pipeline.data.file_list) == 0:
            self.logger.error("Input Error: load data first.")
            return None

        # Get the row numbers of the selected indexes (if any)
        selected_rows = np.unique([index.row() for index in self.table_view.selectedIndexes()])

        # remove these sinusoids from the model
        self.pipeline_thread.start_function('delete_sinusoids', selected_rows)

        return None

    def show_settings_dialog(self):
        """Show a 'settings' dialog with configuration for the application."""
        dialog = gui_config.SettingsDialog(parent=self)

        if dialog.exec():
            # Update any dependent GUI components with new configuration values
            screen = QApplication.primaryScreen()
            screen_size = screen.availableSize()
            h_size = int(screen_size.width() * config.h_size_frac)  # some fraction of the screen width
            v_size = int(screen_size.height() * config.v_size_frac)  # some fraction of the screen height
            self.setGeometry(100, 50, h_size, v_size)

            # update the displayed information
            self.update_info_fields()

        return None

    def export_settings(self):
        """Export the configuration file to a user specified location."""
        suggested_path = os.path.join(config.save_dir, "config.yaml")
        file_path, _ = QFileDialog.getSaveFileName(self, caption="Export Settings", dir=suggested_path,
                                                   filter="YAML Files (*.yaml);;All Files (*)")

        main.save_config(file_path)

        return None

    def import_settings(self):
        """Import the configuration from a user specified file."""
        file_path, _ = QFileDialog.getOpenFileName(self, caption="Import Settings", dir=config.save_dir,
                                                    filter="YAML Files (*.yaml);;All Files (*)")

        main.update_config(file_name=file_path)

        return None

    def show_about_dialog(self):
        """Show an 'about' dialog with information about the application."""
        version = hlp.get_version()
        message = (f"STAR SHINE version {version}\n"
                   "Satellite Time-series Analysis Routine "
                   "using Sinusoids and Harmonics through Iterative Non-linear Extraction\n"
                   "Repository: https://github.com/LucIJspeert/star_shine\n"
                   "Code written by: Luc IJspeert")
        QMessageBox.about(self, "About", message)

        return None


class InputDialog(QDialog):
    def __init__(self, title, *texts):
        super().__init__()
        self.setWindowTitle(title)
        layout = QFormLayout()

        # List to store the line edits
        self.line_edits = []

        # Create a label and line edit for each provided text argument
        for i, text in enumerate(texts):
            label = QLabel(text)
            line_edit = QLineEdit()
            layout.addRow(label, line_edit)
            self.line_edits.append(line_edit)

        # Create a button to accept input
        ok_button = QPushButton("Accept")
        ok_button.clicked.connect(self.accept)
        layout.addWidget(ok_button)

        self.setLayout(layout)

    def get_values(self):
        return [line_edit.text() for line_edit in self.line_edits]


def launch_gui():
    """Launch the Star Shine GUI."""
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    launch_gui()



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains plotting functions for the graphical user interface.

Code written by: Luc IJspeert
"""
import os
import matplotlib as mpl
import matplotlib.figure
import numpy as np
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas, NavigationToolbar2QT

from PySide6.QtWidgets import QWidget, QVBoxLayout
from PySide6.QtCore import Signal

from star_shine.config.helpers import get_images_path


class PlotToolbar(NavigationToolbar2QT):
    """New plot toolbar"""

    click_icon_file = os.path.join(get_images_path(), 'click')
    residual_icon_file = os.path.join(get_images_path(), 'residual')

    # list of toolitems to add to the toolbar, format is:
    # (
    #   text, # the text of the button (often not visible to users)
    #   tooltip_text, # the tooltip shown on hover (where possible)
    #   image_file, # name of the image for the button (without the extension)
    #   name_of_method, # name of the method in NavigationToolbar2 to call
    # )
    NavigationToolbar2QT.toolitems = [
        ('Home', 'Reset original view', 'home', 'home'),
        ('Back', 'Back to previous view', 'back', 'back'),
        ('Forward', 'Forward to next view', 'forward', 'forward'),
        (None, None, None, None),
        ('Pan', 'Left button pans, Right button zooms\nx/y fixes axis, CTRL fixes aspect', 'move', 'pan'),
        ('Zoom', 'Zoom to rectangle\nx/y fixes axis', 'zoom_to_rect', 'zoom'),
        ('Click', 'Click on the plot to interact', click_icon_file, 'click'),
        (None, None, None, None),
        ('Residual', 'Show residual plot', residual_icon_file, 'residual'),
        ('Subplots', 'Configure subplots', 'subplots', 'configure_subplots'),
        ('Customize', 'Edit axis, curve and image parameters', 'qt4_editor_options', 'edit_parameters'),
        (None, None, None, None),
        ('Save', 'Save the figure', 'filesave', 'save_figure')
    ]

    def __init__(self, canvas, parent):
        super().__init__(canvas, parent)
        self._setup_click_button()
        self._setup_residual_button()

    def _setup_click_button(self):
        """Set up the click mode button."""
        for action in self.actions():
            if action.text() == "Click":
                action.setCheckable(True)
                self.click_action = action
                break

        return None

    def click(self, *args):
        """Toggle click mode."""
        if args and args[0]:
            self.click_action.setChecked(True)
            return None
        elif args and not args[0]:
            self.click_action.setChecked(False)
            return None

        # handle state of built-in methods
        if not args and self.mode == 'pan/zoom':
            NavigationToolbar2QT.pan(self, False)
        elif not args and self.mode == 'zoom rect':
            NavigationToolbar2QT.zoom(self, False)

        return None

    def pan(self, *args):
        """Toggle pan mode."""
        # handle state of click button
        if self.click_action.isChecked():
            self.click(False)

        super().pan(*args)

    def zoom(self, *args):
        """Toggle zoom mode."""
        # handle state of click button
        if self.click_action.isChecked():
            self.click(False)

        super().zoom(*args)

    def _setup_residual_button(self):
        """Set up the residual mode button."""
        for action in self.actions():
            if action.text() == "Residual":
                action.setCheckable(True)
                self.residual_action = action
                break

        return None

    def residual(self, *args):
        """Toggle residual mode."""
        if args and args[0]:
            self.residual_action.setChecked(True)
            return None
        elif args and not args[0]:
            self.residual_action.setChecked(False)
            return None

        return None


class PlotWidget(QWidget):
    """A widget for displaying plots using Matplotlib in a Qt application.

    Attributes
    ----------
    """
    # Define a signal that emits the plot ID and clicked coordinates
    click_signal = Signal(float, float, int)
    residual_signal = Signal()

    def __init__(self, title='Plot', xlabel='x', ylabel='y'):
        """A widget for displaying plots using Matplotlib in a Qt application.

        Parameters
        ----------
        title: str, optional
            Title of the plot. Default is 'Plot'.
        xlabel: str, optional
            Label for the x-axis. Default is 'x'.
        ylabel: str, optional
            Label for the y-axis. Default is 'y'.
        """
        super().__init__()
        # store some info
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel

        # set up the figure and canvas with an axis
        self.figure = mpl.figure.Figure()
        self.canvas = FigureCanvas(self.figure)
        # self.figure.patch.set_facecolor('grey')
        # self.figure.patch.set_alpha(0.0)

        # Add toolbar for interactivity
        self.toolbar = PlotToolbar(self.canvas, self)
        self.show_residual = False

        # make an axis and set the labels
        self.ax = self.figure.add_subplot(111)
        self._set_labels()

        # make the layout and add the toolbar and canvas widgets
        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        self.setLayout(layout)

        # Connect the mouse click event to a custom method
        self.canvas.mpl_connect('button_press_event', self.on_click)

        # Connect the residual action to the on_residual method
        self.toolbar.residual_action.triggered.connect(self.on_residual)

        # plot types and properties supported
        self.plot_type_list = ['plot', 'scatter', 'vlines']
        self.plot_property_list = ['_xs', '_ys', '_labels', '_colors']
        self.property_fill_values = {'_xs': [], '_ys': [], '_labels': '', '_colors': None}

        # make references for the plot data and plot elements
        for plot_type in self.plot_type_list:
            setattr(self, plot_type + '_art', [])
            for plot_property in self.plot_property_list:
                key = plot_type + plot_property
                setattr(self, key, [])

    def _set_labels(self):
        """Set the axes labels and title."""
        self.ax.set_title(self.title)
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)

        return None

    def clear_plot(self):
        """Clear the plot"""
        self.ax.clear()

        # re-apply some elements
        self._set_labels()

        return None

    def on_click(self, event):
        """Click event"""
        # Left mouse button click
        if self.toolbar.click_action.isChecked() and event.button == 1:
            x, y = event.xdata, event.ydata
            # Ensure valid coordinates
            if x is not None and y is not None:
                self.click_signal.emit(x, y, event.button)

        # Right mouse button click
        if self.toolbar.click_action.isChecked() and event.button == 3:
            x, y = event.xdata, event.ydata
            # Ensure valid coordinates
            if x is not None and y is not None:
                self.click_signal.emit(x, y, event.button)

        return None

    def on_residual(self, event):
        """Residual event"""
        self.show_residual = self.toolbar.residual_action.isChecked()
        self.residual_signal.emit()

        return None

    def set_plot_data(self, **kwargs):
        """Set the plot data, wiping existing data.

        Keywords:
        line_xs, line_ys, line_labels, line_colors,
        scatter_xs, scatter_ys, scatter_labels, scatter_colors,
        vlines_xs, vlines_ys, vlines_labels, vlines_colors

        All arguments must be supplied as lists. Data is only set if `<type>_xs` and `<type>_ys`
        for the plot type is given. Lists per plot type are assumed of equal length.
        """
        # see if there is data in kwargs for any of the plot types
        for plot_type in self.plot_type_list:
            # go through the supplied and existing plot properties to check lengths
            n_existing = 0
            n_supplied = 0
            for plot_property in self.plot_property_list:
                key = plot_type + plot_property
                n_existing = max(n_existing, len(getattr(self, key)))  #  take the max length
                if key in kwargs.keys():
                    n_supplied = max(n_supplied, len(kwargs[key]))  #  take the max length

            # positive if more data supplied than existing plots
            n_diff = n_supplied - n_existing

            # go through the plot properties to set them accordingly
            for plot_property in self.plot_property_list:
                key = plot_type + plot_property
                if key in kwargs.keys() and (len(kwargs[key]) == n_supplied):
                    # set the plot property to the supplied value(s)
                    setattr(self, key, kwargs[key])
                elif key in kwargs.keys():
                    # set the plot property to the supplied value(s) and extend them with fill values
                    setattr(self, key, kwargs[key])
                    n_diff_i = n_supplied - len(kwargs[key])
                    getattr(self, key).extend([self.property_fill_values[plot_property] for _ in range(n_diff_i)])
                elif n_existing == 0:
                    # set the plot property to its fill value(s)
                    setattr(self, key, [self.property_fill_values[plot_property] for _ in range(n_supplied)])
                elif n_existing < n_supplied:
                    # extend the plot property with its fill value(s)
                    getattr(self, key).extend([self.property_fill_values[plot_property] for _ in range(n_diff)])
                else:  # n_existing >= n_supplied
                    # remove excess plot properties
                    setattr(self, key, getattr(self, key)[:n_supplied])

        return None

    @staticmethod
    def _update_plot_element(plot_type, plot_element, **kwargs):
        """Update an existing plot element."""
        # update the data
        if plot_type == 'plot':
            plot_element.set_xdata(kwargs['x'])
            plot_element.set_ydata(kwargs['y'])
        elif plot_type == 'scatter':
            plot_element.set_offsets(np.column_stack((kwargs['x'], kwargs['y'])))
        else:  # plot_type == 'vlines'
            plot_element.set_segments([[[xi, 0], [xi, yi]] for xi, yi in zip(kwargs['x'], kwargs['y'])])

        # update colour
        if 'color' in kwargs and kwargs['color'] is not None:
            if plot_type == 'plot':
                plot_element.set_color(kwargs['color'])
            elif plot_type == 'scatter':
                plot_element.set_facecolors(kwargs['color'])
            else:  # plot_type == 'vlines'
                plot_element.set_colors(kwargs['color'])

        # update label
        if 'label' in kwargs and kwargs['label'] != '':
            plot_element.set_label(kwargs['label'])

        return

    def _create_plot_element(self, plot_type, **kwargs):
        """Create a new plot element."""
        # make sure mpl gets the right keywords
        mpl_kwargs = {}
        if 'color' in kwargs and kwargs['color'] is not None:
            if plot_type == 'plot':
                mpl_kwargs['color'] = kwargs['color']
            elif plot_type == 'scatter':
                mpl_kwargs['c'] = kwargs['color']
            else:  # plot_type == 'vlines'
                mpl_kwargs['colors'] = kwargs['color']

        if 'label' in kwargs and kwargs['label'] != '':
            mpl_kwargs['label'] = kwargs['label']

        if plot_type == 'plot':
            art = self.ax.plot(kwargs['x'], kwargs['y'], **mpl_kwargs)[0]  # returns list
        elif plot_type == 'scatter':
            art = self.ax.scatter(kwargs['x'], kwargs['y'], marker='.', **mpl_kwargs)
        else:  # plot_type == 'vlines'
            ymin = np.zeros_like(kwargs['x'])
            art = self.ax.vlines(kwargs['x'], ymin=ymin, ymax=kwargs['y'], **mpl_kwargs)

        return art

    def update_plot(self):
        """Update the plot in the widget."""
        # update the plot (with altered or appended data)
        for plot_type in self.plot_type_list:
            plot_elements = getattr(self, plot_type + '_art')
            xs = getattr(self, plot_type + '_xs')
            ys = getattr(self, plot_type + '_ys')
            colors = getattr(self, plot_type + '_colors')
            labels = getattr(self, plot_type + '_labels')

            # for each plot element: if the element exists, update it, else create it
            for i in range(len(xs)):
                kwargs = {'x': xs[i], 'y': ys[i], 'color': colors[i], 'label': labels[i]}

                if i < len(plot_elements):
                    # update the plot element
                    self._update_plot_element(plot_type, plot_elements[i], **kwargs)
                else:
                    # create new plot element
                    art = self._create_plot_element(plot_type, **kwargs)
                    plot_elements.append(art)

        # Redraw the canvas to reflect changes
        # self.ax.legend()
        self.figure.tight_layout()  # needs to happen before draw
        self.canvas.draw()

        return None

    def new_plot(self):
        """Plot the data in the widget."""
        # Start by clearing the canvas
        self.clear_plot()
        self.plot_art = []  # reset all plot elements
        self.scatter_art = []  # reset all scatter elements
        self.vlines_art = []  # reset all vlines elements

        # Plot the line plot(s)
        self.update_plot()

        return None



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains functions for goodness of fit determination.

Code written by: Luc IJspeert
"""

import numpy as np
import scipy as sp
import numba as nb

from star_shine.core import periodogram as pdg


@nb.njit(cache=True)
def calc_iid_normal_likelihood(residuals):
    """Natural logarithm of the independent and identically distributed likelihood function.

    Under the assumption that the errors are independent and identically distributed
    according to a normal distribution, the likelihood becomes:
    ln(L(θ)) = -n/2 (ln(2 pi σ^2) + 1)
    and σ^2 is estimated as σ^2 = sum((residuals)^2)/n

    Parameters
    ----------
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model

    Returns
    -------
    float
        Natural logarithm of the likelihood
    """
    n = len(residuals)
    # like = -n / 2 * (np.log(2 * np.pi * np.sum(residuals**2) / n) + 1)
    # originally un-JIT-ted function, but for loop is quicker with numba
    sum_r_2 = 0
    for i in range(n):
        sum_r_2 += residuals[i] ** 2

    like = -n / 2 * (np.log(2 * np.pi * sum_r_2 / n) + 1)

    return like


def calc_approx_did_likelihood(time, residuals):
    """Approximation for the likelihood using periodograms.

    This function approximates the dependent likelihood for correlated data.
    ln(L(θ)) =  -n ln(2 pi) - sum(ln(PSD(residuals)))

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model

    Returns
    -------
    float
        Log-likelihood approximation
    """
    n = len(time)

    # Compute the Lomb-Scargle periodogram of the data
    freqs, psd = pdg.scargle_parallel(time, residuals, f0=0, norm='psd')  # automatically mean subtracted

    # Compute the Whittle likelihood
    like = -n * np.log(2 * np.pi) - np.sum(np.log(psd))

    return like


def calc_whittle_likelihood(time, flux, model):
    """Whittle likelihood approximation using periodograms.

    This function approximates the dependent likelihood for correlated data.
    It assumes the data is identically distributed according to a normal
    distribution.
    ln(L(θ)) =  -n ln(2 pi) - sum(ln(PSD_model) + (PSD_data / PSD_model))

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    model: numpy.ndarray[Any, dtype[float]]
        Model values of the time series

    Returns
    -------
    float
        Log-likelihood approximation
    """
    n = len(time)

    # Compute the Lomb-Scargle periodogram of the data
    freqs, psd_d = pdg.scargle_parallel(time, flux, f0=0, norm='psd')  # automatically mean subtracted

    # Compute the Lomb-Scargle periodogram of the model
    freqs_m, psd_m = pdg.scargle_parallel(time, model, f0=0, norm='psd')  # automatically mean subtracted

    # Avoid division by zero in likelihood calculation
    psd_m = np.maximum(psd_m, 1e-15)  # Ensure numerical stability

    # Compute the Whittle likelihood
    like = -n * np.log(2 * np.pi) - np.sum(np.log(psd_m) + (psd_d / psd_m))

    return like


def calc_did_normal_likelihood(time, residuals):
    """Natural logarithm of the dependent and identically distributed likelihood function.

    Correlation in the data is taken into account. The data is still assumed to be
    identically distributed according to a normal distribution.

    ln(L(θ)) = -n ln(2 pi) / 2 - ln(det(∑)) / 2 - residuals @ ∑^-1 @ residuals^T / 2
    ∑ is the covariance matrix

    The covariance matrix is calculated using the power spectral density, following
    the Wiener–Khinchin theorem.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model

    Returns
    -------
    float
        Natural logarithm of the likelihood
    """
    n = len(residuals)

    # calculate the PSD, fast
    freqs, psd = pdg.scargle_parallel(time, residuals, f0=0, norm='psd')

    # calculate the autocorrelation function
    psd_ext = np.append(psd, psd[-1:0:-1])  # double the PSD domain for ifft
    acf = np.fft.ifft(psd_ext)

    # unbias the variance measure and put the array the right way around
    acf = np.real(np.append(acf[len(freqs):], acf[:len(freqs)])) * n / (n - 1)

    # calculate the acf lags
    lags = np.fft.fftfreq(len(psd_ext), d=(freqs[1] - freqs[0]))
    lags = np.append(lags[len(psd):], lags[:len(psd)])  # put them the right way around

    # make the lags matrix, but re-use the same matrix
    matrix = time - time[:, np.newaxis]  # lags_matrix, same as np.outer

    # interpolate - I need the lags at specific times
    matrix = np.interp(matrix, lags, acf)  # cov_matrix, already mean-subtracted in PSD

    # Compute the Cholesky decomposition of cov_matrix (by definition positive definite)
    matrix = sp.linalg.cho_factor(matrix, lower=False, overwrite_a=True, check_finite=False)  # cho_decomp

    # Solve M @ x = v^T using the Cholesky factorization (x = M^-1 v^T)
    x = sp.linalg.cho_solve(matrix, residuals[:, np.newaxis], check_finite=False)

    # log of the exponent - analogous to the matrix multiplication
    ln_exp = (residuals @ x)[0]  # v @ x = v @ M^-1 @ v^T

    # log of the determinant (avoids too small eigenvalues that would result in 0)
    ln_det = 2 * np.sum(np.log(np.diag(matrix[0])))

    # likelihood for multivariate normal distribution
    like = -n * np.log(2 * np.pi) / 2 - ln_det / 2 - ln_exp / 2

    return like


def calc_ddd_normal_likelihood(time, residuals, flux_err):
    """Natural logarithm of the dependent and differently distributed likelihood function.

    Only assumes that the data is distributed according to a normal distribution.
    Correlation in the data is taken into account. The measurement errors take
    precedence over the measured variance in the data. This means the distributions
    need not be identical, either.

    ln(L(θ)) = -n ln(2 pi) / 2 - ln(det(∑)) / 2 - residuals @ ∑^-1 @ residuals^T / 2
    ∑ is the covariance matrix

    The covariance matrix is calculated using the power spectral density, following
    the Wiener–Khinchin theorem.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values

    Returns
    -------
    float
        Natural logarithm of the likelihood
    """
    n = len(residuals)

    # calculate the PSD, fast
    freqs, psd = pdg.scargle_parallel(time, residuals, f0=0, norm='psd')

    # calculate the autocorrelation function
    psd_ext = np.append(psd, psd[-1:0:-1])  # double the PSD domain for ifft
    acf = np.fft.ifft(psd_ext)

    # unbias the variance measure and put the array the right way around
    acf = np.real(np.append(acf[len(freqs):], acf[:len(freqs)])) * n / (n - 1)

    # calculate the acf lags
    lags = np.fft.fftfreq(len(psd_ext), d=(freqs[1] - freqs[0]))
    lags = np.append(lags[len(psd):], lags[:len(psd)])  # put them the right way around

    # make the lags matrix, but re-use the same matrix
    matrix = time - time[:, np.newaxis]  # lags_matrix, same as np.outer

    # interpolate - I need the lags at specific times
    matrix = np.interp(matrix, lags, acf)  # cov_matrix, already mean-subtracted in PSD

    # substitute individual data errors if given
    var = matrix[0, 0]  # diag elements are the same by construction
    corr_matrix = matrix / var  # divide out the variance to get correlation matrix
    err_matrix = flux_err * flux_err[:, np.newaxis]  # make matrix of measurement errors (same as np.outer)
    matrix = err_matrix * corr_matrix  # multiply to get back to covariance

    # Compute the Cholesky decomposition of cov_matrix (by definition positive definite)
    matrix = sp.linalg.cho_factor(matrix, lower=False, overwrite_a=True, check_finite=False)  # cho_decomp

    # Solve M @ x = v^T using the Cholesky factorization (x = M^-1 v^T)
    x = sp.linalg.cho_solve(matrix, residuals[:, np.newaxis], check_finite=False)

    # log of the exponent - analogous to the matrix multiplication
    ln_exp = (residuals @ x)[0]  # v @ x = v @ M^-1 @ v^T

    # log of the determinant (avoids too small eigenvalues that would result in 0)
    ln_det = 2 * np.sum(np.log(np.diag(matrix[0])))

    # likelihood for multivariate normal distribution
    like = -n * np.log(2 * np.pi) / 2 - ln_det / 2 - ln_exp / 2

    return like


@nb.njit(cache=True)
def calc_bic(residuals, n_param):
    """Bayesian Information Criterion.

    Parameters
    ----------
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model
    n_param: int
        Number of free parameters in the model

    Returns
    -------
    float
        Bayesian Information Criterion

    Notes
    -----
    BIC = −2 ln(L(θ)) + k ln(n)
    where L is the likelihood as function of the parameters θ, n the number of data points
    and k the number of free parameters.

    Under the assumption that the errors are independent and identically distributed
    according to a normal distribution, the likelihood becomes:
    ln(L(θ)) = -n/2 (ln(2 pi σ^2) + 1)
    and σ^2 is the error variance estimated as σ^2 = sum((residuals)^2)/n
    (residuals being data - model).

    Combining this gives:
    BIC = n ln(2 pi σ^2) + n + k ln(n)
    """
    n = len(residuals)

    sum_r_2 = 0
    for i in range(n):
        sum_r_2 += residuals[i]**2

    bic = n * np.log(2 * np.pi * sum_r_2 / n) + n + n_param * np.log(n)

    return bic



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains classes for handling the time series, including combining it with the full model.

Code written by: Luc IJspeert
"""
import numpy as np

from star_shine.core import model as mdl, goodness_of_fit as gof, periodogram as pdg
from star_shine.config import data_properties as dp


class TimeSeries:
    """This class handles the time series.

    Attributes
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series.
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    n_time: int
        Number of data points in the time series.
    n_chunks: int
        Number of time chunks in the time series.
    t_tot: float
        Total time base of observations.
    t_mean: float
        Time reference (zero) point of the full light curve.
    t_mean_chunk: numpy.ndarray[Any, dtype[float]]
        Time reference (zero) point per chunk.
    t_step: float
        Median time step of observations.
    pd_f0: float
        Starting frequency of the periodogram.
    pd_fn: float
        Last frequency of the periodogram.
    pd_df: float
        Frequency sampling space of the periodogram
    pd_freqs: numpy.ndarray[Any, dtype[float]]
        Frequencies at which the periodogram was calculated
    pd_ampls: numpy.ndarray[Any, dtype[float]]
        The periodogram amplitudes.
    f_resolution: float
        Frequency resolution of the time series
    snr_threshold: float
        Signal-to-noise threshold for this data set.
    """

    def __init__(self, time, flux, flux_err, i_chunks):
        """Initialises the Result object.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        flux: numpy.ndarray[Any, dtype[float]]
            Measurement values of the time series.
        flux_err: numpy.ndarray[Any, dtype[float]]
            Errors in the measurement values.
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        """
        # time series
        self.time = time
        self.flux = flux
        self.flux_err = flux_err
        self.i_chunks = i_chunks

        # some numbers
        self.n_time = len(time)
        self.n_chunks = len(i_chunks)

        # set time properties
        self.t_tot = np.ptp(self.time)
        self.t_mean = np.mean(self.time)
        self.t_mean_chunk = np.array([np.mean(self.time[ch[0]:ch[1]]) for ch in self.i_chunks])
        self.t_step = np.median(np.diff(self.time))

        # settings for periodograms
        self.pd_f0 = 0.01 / self.t_tot  # lower than T/100 no good
        self.pd_df = 0.1 / self.t_tot  # default frequency sampling is about 1/10 of frequency resolution
        self.pd_fn = 0.  # set by update_properties

        # other properties that rely on config
        self.f_resolution = 0.
        self.snr_threshold = 0.
        self.update_properties()  # update these with a function

        # periodogram
        out = pdg.scargle_parallel(self.time, self.flux, f0=self.pd_f0, fn=self.pd_fn, df=self.pd_df, norm='amplitude')
        self.pd_freqs = out[0]
        self.pd_ampls = out[1]

    def update_properties(self):
        """Calculate the properties of the data and fill them in.

        Running this function again will re-evaluate some properties, for if the configuration changed.
        """
        # set data properties relying on config
        self.pd_fn = dp.nyquist_frequency(self.time)
        self.f_resolution = dp.frequency_resolution(self.time)
        self.snr_threshold = dp.signal_to_noise_threshold(self.time)

        return None


class TimeSeriesModel(TimeSeries):
    """This class handles the full time series model.

    Attributes
    ----------
    linear: star_shine.core.time_series.LinearModel
        Model of the piece-wise linear curve.
    sinusoid: star_shine.core.time_series.SinusoidModel
        Model of the sinusoids.
    """

    def __init__(self, time, flux, flux_err, i_chunks):
        """Initialises the Result object.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        flux: numpy.ndarray[Any, dtype[float]]
            Measurement values of the time series.
        flux_err: numpy.ndarray[Any, dtype[float]]
            Errors in the measurement values.
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        """
        # instantiate time series
        super().__init__(time, flux, flux_err, i_chunks)

        # time series models making up the full model
        self.linear = mdl.LinearModel(self.n_time)
        self.sinusoid = mdl.SinusoidModel(self.n_time)

    @staticmethod
    def from_time_series(time_series):
        """Create a TimeSeriesModel instance from a TimeSeries instance."""
        return TimeSeriesModel(time=time_series.time, flux=time_series.flux, flux_err=time_series.flux_err,
                               i_chunks=time_series.i_chunks)

    @property
    def n_param(self):
        """Return the number of parameters of the time series model.

        Returns
        -------
        int
            Number of free parameters in the model.
        """
        return self.linear.n_param + self.sinusoid.n_param

    def get_parameters(self):
        """Get the current model parameters.

        Returns
        -------
        tuple
            Consisting of five numpy.ndarray[Any, dtype[float]] for const, slope, f_n, a_n, ph_n.
        """
        return *self.linear.get_linear_parameters(), *self.sinusoid.get_sinusoid_parameters()

    def model(self):
        """The full time series model.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Combined time series model.
        """
        return self.linear.linear_model + self.sinusoid.sinusoid_model

    def residual(self):
        """The residual of the flux minus the model.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Flux minus the current time series model.
        """
        return self.flux - self.model()

    def bic(self):
        """Calculate the BIC of the residual.

        Returns
        -------
        float
            BIC of the current time series model.
        """
        return gof.calc_bic(self.residual(), self.n_param)

    def periodogram(self, subtract_model=True):
        """Get the Lomb-Scargle periodogram of the time series.

        Uses the cached values for the original data, if the model is subtracted the periodogram is always recomputed.

        Parameters
        ----------
        subtract_model: bool
            Subtract the time series model from the data.

        Returns
        -------
        tuple
            Contains the frequencies numpy.ndarray[Any, dtype[float]]
            and the spectrum numpy.ndarray[Any, dtype[float]]
        """
        if subtract_model:
            f0, fn, df = self.pd_f0, self.pd_fn, self.pd_df
            f, a = pdg.scargle_parallel(self.time, self.residual(), f0=f0, fn=fn, df=df, norm='amplitude')
        else:
            f, a = self.pd_freqs, self.pd_ampls

        return f, a

    def calc_model(self, indices=None):
        """Calculate the full time series model (disregarding include).

        Parameters
        ----------
        indices: numpy.ndarray[Any, dtype[int]]
            Indices for the sinusoids to include.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Combined time series model.
        """
        model_linear = self.linear.calc_linear_model(self.time, self.i_chunks)
        model_sinusoid = self.sinusoid.calc_sinusoid_model(self.time, indices=indices)

        return model_linear + model_sinusoid

    def calc_periodogram(self):
        """Calculate Lomb-Scargle periodogram of the time series (disregarding include).

        Returns
        -------
        tuple
            Contains the frequencies numpy.ndarray[Any, dtype[float]]
            and the spectrum numpy.ndarray[Any, dtype[float]]
        """
        f0, fn, df = self.pd_f0, self.pd_fn, self.pd_df
        f, a = pdg.scargle_parallel(self.time, self.flux - self.calc_model(), f0=f0, fn=fn, df=df, norm='amplitude')

        return f, a

    def set_linear_model(self, const_new, slope_new):
        """Delegates to set_linear_model of LinearModel."""
        self.linear.set_linear_model(self.time, const_new, slope_new, self.i_chunks)

    def update_linear_model(self):
        """Delegates to update_linear_model of LinearModel."""
        self.linear.update_linear_model(self.time, self.flux - self.sinusoid.sinusoid_model, self.i_chunks)

    def update_linear_uncertainties(self):
        """Delegates to update_linear_model_err of LinearModel."""
        self.linear.update_linear_uncertainties(self.time, self.residual(), self.i_chunks)

    def set_sinusoids(self, *args, **kwargs):
        """Delegates to set_sinusoids of SinusoidModel."""
        self.sinusoid.set_sinusoids(self.time, *args, **kwargs)

    def add_sinusoids(self, *args, **kwargs):
        """Delegates to add_sinusoids of SinusoidModel."""
        self.sinusoid.add_sinusoids(self.time, *args, **kwargs)

    def update_sinusoids(self, *args, **kwargs):
        """Delegates to update_sinusoids of SinusoidModel."""
        self.sinusoid.update_sinusoids(self.time, *args, **kwargs)

    def remove_sinusoids(self, indices):
        """Delegates to remove_sinusoids of SinusoidModel."""
        self.sinusoid.remove_sinusoids(self.time, indices)

    def include_sinusoids(self, indices):
        """Delegates to include_sinusoids of SinusoidModel."""
        self.sinusoid.include_sinusoids(self.time, indices)

    def exclude_sinusoids(self, indices):
        """Delegates to exclude_sinusoids of SinusoidModel."""
        self.sinusoid.exclude_sinusoids(self.time, indices)

    def remove_excluded(self):
        """Delegates to remove_excluded of SinusoidModel."""
        self.sinusoid.remove_excluded()

    def update_sinusoid_uncertainties(self):
        """Delegates to update_sinusoid_uncertainties of SinusoidModel."""
        self.sinusoid.update_sinusoid_uncertainties(self.time, self.residual(), self.flux_err, self.i_chunks)

    def update_sinusoid_uncertainties_harmonic(self):
        """Delegates to update_sinusoid_uncertainties_harmonic of SinusoidModel."""
        self.sinusoid.update_sinusoid_uncertainties_harmonic()

    def update_sinusoid_passing_sigma(self):
        """Delegates to update_sinusoid_passing_sigma of SinusoidModel."""
        self.sinusoid.update_sinusoid_passing_sigma()

    def update_sinusoid_passing_snr(self, window_width=1.):
        """Delegates to update_sinusoid_passing_snr of SinusoidModel."""
        self.sinusoid.update_sinusoid_passing_snr(self.time, self.residual(), window_width=window_width)

    def update_sinusoid_passing_harmonic(self):
        """Delegates to update_sinusoid_passing_harmonic of SinusoidModel."""
        self.sinusoid.update_sinusoid_passing_harmonic(self.f_resolution)
